/* by KI4LKF */

#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>
#include <db.h>

#define CALL_SIZE 8
#define MAXHOSTNAMELEN 64
#define VERSION "2.9"

static DB *dbp = NULL;
static DBT key, data;
static void print_calls();
static void add_call(char *arg, char *host);
static void search_call(char *arg);
static void delete_call(char *arg);

static void print_calls()
{
   int rc;
   DBC *cursorp;
   unsigned long num_calls = 0;

   rc = dbp->cursor(dbp, NULL, &cursorp, 0);
   if (rc != 0)
   {
      fprintf(stderr,"db_cursor() failed: error=%s\n",db_strerror(rc));
      return;
   }

   memset(&key, 0, sizeof(DBT));
   memset(&data, 0, sizeof(DBT));

   while ((rc = cursorp->c_get(cursorp, &key, &data, DB_NEXT)) == 0) 
   {
      printf("[%s] [%s] \n", (char *)key.data, (char *)data.data);
      num_calls ++;
   }
   if (rc != DB_NOTFOUND)
      fprintf(stderr,"db_cursor() failed: error=%s\n",db_strerror(rc));
   
   if (cursorp != NULL)
      cursorp->c_close(cursorp);

   return;

}

static void search_call(char *arg)
{
   int rc;
   char call[CALL_SIZE + 1];
   char payload[MAXHOSTNAMELEN + 1];

   memset(&key, 0, sizeof(DBT));
   memset(&data, 0, sizeof(DBT));

   memset(call, ' ', CALL_SIZE);
   call[CALL_SIZE] = '\0';

   memcpy(call, arg, strlen(arg));

   key.data = call;
   key.size = CALL_SIZE + 1;

   data.data = payload;
   data.size = MAXHOSTNAMELEN + 1;

   rc = dbp->get(dbp, NULL, &key, &data, 0);
   if (rc != 0)
      fprintf(stderr,"db_get() failed: call=%s, error=%s\n",call, db_strerror(rc));
   else
      fprintf(stderr,"[%s],[%s]\n", (char *)key.data, (char *)data.data);

   return;
}

static void delete_call(char *arg)
{
   int rc;
   char call[CALL_SIZE + 1];

   memset(&key, 0, sizeof(DBT));

   memset(call, ' ', CALL_SIZE);
   call[CALL_SIZE] = '\0';

   memcpy(call, arg, strlen(arg));

   key.data = call;
   key.size = CALL_SIZE + 1;

   rc = dbp->del(dbp, NULL, &key, 0);
   if (rc != 0)
      fprintf(stderr,"db_del() failed: call=%s, error=%s\n",call, db_strerror(rc));
   else
      fprintf(stderr,"Call %s deleted OK\n", call);

   return;
}

static void add_call(char *arg, char *host)
{
   int rc;
   char call[CALL_SIZE + 1];

   memset(&key, 0, sizeof(DBT));
   memset(&data, 0, sizeof(DBT));

   memset(call, ' ', CALL_SIZE);
   call[CALL_SIZE] = '\0';

   memcpy(call, arg, strlen(arg));
   key.data = call;
   key.size = CALL_SIZE + 1;

   data.data = host;
   data.size = strlen(host) + 1;

   rc = dbp->put(dbp, NULL, &key, &data, DB_NOOVERWRITE);
   if (rc != 0)
      fprintf(stderr,"db_put() failed: call=%s, error=%s\n",call, db_strerror(rc));
   else 
      fprintf(stderr, "Added [%s] OK\n", call);
  
   return;
}

static void usage()
{
   fprintf(stderr,"\nTo search for a callsign:\n");
   fprintf(stderr,"dextra_access_db <your dextra database> s <callsign>\n");
   fprintf(stderr,"    Example:  dextra_access_db dextra.db s DB0SAT\n");
   fprintf(stderr,"    Example:  dextra_access_db dextra.db s DH2YBE\n");
   fprintf(stderr,"    Example:  dextra_access_db dextra.db s \"DH2YBE D\"\n\n");

   fprintf(stderr,"To delete a callsign:\n");
   fprintf(stderr,"dextra_access_db <your dextra database> d <callsign>\n");
   fprintf(stderr,"    Example: dextra_access_db dextra.db d DB0SAT\n");
   fprintf(stderr,"    Example: dextra_access_db dextra.db d DH2YBE\n"); 
   fprintf(stderr,"    Example: dextra_access_db dextra.db d \"DH2YBE D\"\n\n");
  
   fprintf(stderr,"To add a callsign:\n"); 
   fprintf(stderr,"dextra_access_db <your dextra database> a <callsign> [hostname]\n");
   fprintf(stderr,"    Example: dextra_access_db dextra.db a DG1HT\n");
   fprintf(stderr,"    Example: dextra_access_db dextra.db a \"DG1HT  D\"\n");
   fprintf(stderr,"    Example: dextra_access_db dextra.db a XRF010 \"d-star.ath.cx\"\n\n");

   fprintf(stderr,"To print the database\n");
   fprintf(stderr,"dextra_access_db <your dextra database> p\n");
   fprintf(stderr,"    Example: dextra_access_db dextra.db p\n");

   fprintf(stderr,"dextra_access_db version %s\n", VERSION);
   return;
}

int main(int argc, char *argv[])
{
   int rc = 0;
   u_int32_t flags = 0;

   if (argc < 3)
   {
      usage();
      return 1;
   }

   fprintf(stderr,"Trying to open database %s\n", argv[1]);
   rc = db_create(&dbp, NULL, 0);
   if (rc != 0) 
   {
      fprintf(stderr,"db_create() failed: error=%s\n", db_strerror(rc));
      return 1;
   } 

   rc = dbp->open(dbp,NULL,argv[1],NULL,DB_BTREE,flags,0);
   if (rc != 0)
   {
      fprintf(stderr,"db_open() failed: error=%s\n", db_strerror(rc));
      return 1;
   }
   fprintf(stderr,"Database %s opened OK\n", argv[1]);

   if (argv[2][0] == 'p')
      print_calls();
   else
   if (argv[2][0] == 'a')
   {
      if (argc < 4)
         usage();
      else
      if (argc == 4)
         add_call(argv[3], (char *)"");
      else
      if (argc == 5)
         add_call(argv[3], argv[4]);
   }
   else
   if (argv[2][0] == 's')
   {
      if (argc != 4)
         usage();
      else
         search_call(argv[3]);
   }
   else
   if (argv[2][0] == 'd')
   {
      if (argc != 4)
         usage();
      else
         delete_call(argv[3]);
   }

   fprintf(stderr,"Trying to close database %s\n", argv[1]);
   if (dbp != NULL)
   {
      dbp->close(dbp, 0);
      fprintf(stderr,"Closed database %s\n", argv[1]);
   }
   return 0;
}

