
/* by KI4LKF */

#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <limits.h>
#include <db.h>

#define CALL_SIZE 8
#define MAXHOSTNAMELEN 64
#define VERSION "3.6"

static DB *dbp = NULL;

static bool search_call(char *call_module, char *call)
{
   int rc;
   DBT key, data;
   char payload[MAXHOSTNAMELEN + 1];
   bool ok = false;

   memset(&key, 0, sizeof(DBT));
   memset(&data, 0, sizeof(DBT));

   key.data = call_module;
   key.size = CALL_SIZE + 1;

   data.data = payload;
   data.size = MAXHOSTNAMELEN + 1;

   rc = dbp->get(dbp, NULL, &key, &data, 0);
   if (rc != 0)
      ok = false;
   else
   {
      if (*(char *)data.data == '\0')
         ok = false;
      else
      {
         printf("User %s will NOT be added, already exists as: [%s],[%s]\n", call, (char *)key.data, (char *)data.data);
         ok = true;
      }
   }
   return ok;
}

int main(int argc, char *argv[])
{
   int rc = 0;
   u_int32_t flags = 0;
   DBT key, data;

   FILE *fp = NULL;
   char inbuf[1024];
   char *p = NULL;

   const char *delim = "|";

   char *tok;
   char call[CALL_SIZE + 1];
   char payload[MAXHOSTNAMELEN + 1];
   char call_module[CALL_SIZE + 1];
   unsigned long calls_added = 0;

   if (argc != 3)
   {
      printf("dextra_create_db v%s\n", VERSION);
      printf("\nUsage: dextra_create_db <your new dextra database> <inputFile>\n");
      printf("Example: dextra_create_db dextra.db users.txt\n\n");
      printf("   dextra.db database will be created anew and populated with\n");
      printf("      callsigns listed in the file users.txt...\n");
      printf("   The format of the file users.txt is a plain text file\n\n");
      printf("   Example:(The 1st line is a reflector, the 2nd line is a repeater, the 3rd line is a radio user.\n");
      printf("      The reflectors are always first in the list, then the repeaters, then the users.\n");
      printf("                   XRF010  |xrf010.xreflector.net\n");
      printf("                   KJ4NHF  |1.2.3.4\n");
      printf("                   KI4LKF\n");
      printf("  Reflectors or repeaters MUST have a hostname or an IP address and the 2 fields must be separated by a vertical line\n");
      printf("  Radio users must NOT have data next to their callsign\n");
      printf("  XRF reflector callsigns must be exactly 6 characters long and must start with XRF followed by a 3-digit code\n");
      printf("  Repeater callsigns are up to 6 characters long.\n");
      printf("  User callsigns are up to 8 characters long.\n");
      return 1;
   }

   printf("Trying to open file %s\n", argv[2]);
   fp = fopen(argv[2], "r");
   if (fp == NULL)
   {
      printf("Failed to open file %s\n", argv[2]);
      return 1;
   }
   printf("Opened file %s OK\n", argv[2]);

   printf("Trying to initialize new database %s\n", argv[1]);
   rc = db_create(&dbp, NULL, 0);
   if (rc != 0) 
   {
      printf("db_create() failed: error=%s\n", db_strerror(rc));
      fclose(fp);
      return 1;
   } 

   flags = DB_CREATE | DB_EXCL;
   rc = dbp->open(dbp,NULL,argv[1],NULL,DB_BTREE,flags,0);
   if (rc != 0)
   {
      printf("db_open() failed: error=%s\n", db_strerror(rc));
      fclose(fp);
      return 1;
   }
   printf("Database %s initialized OK\n", argv[1]);
   
   printf("Populating database %s from input file %s\n", argv[1], argv[2]);  

   while (fgets(inbuf, 1020, fp) != NULL)
   {
      p = strchr(inbuf, '\r');
      if (p)
         *p = '\0';

      p = strchr(inbuf, '\n');
      if (p)
          *p = '\0';

      memset(&key, 0, sizeof(DBT));
      memset(&data, 0, sizeof(DBT));

      tok = strtok(inbuf, delim); 
      if (!tok)
         continue;  /* Invalid input line */

      /* tok contains the callsign */
      if (strlen(tok) > CALL_SIZE)
      {
         printf("Invalid call %s\n", tok);
         continue;
      }
      if (strlen(tok) < 1)
      {
         printf("Invalid call %s\n", tok);
         continue;
      }

      memset(call, ' ', CALL_SIZE);
      call[CALL_SIZE] = '\0';
      memcpy(call, tok, strlen(tok));
         
      key.data = call;
      key.size = CALL_SIZE + 1;

      tok = strtok(NULL, delim);
      if (!tok)
      {
         /* we are trying to add a USER, but maybe this is an entry for a repeater module */
         if (call[7] != ' ')
         {
            strcpy(call_module, call);
            call_module[7] = ' ';
            if (search_call(call_module,call))
               continue;
         }
         payload[0] = '\0';
         data.data = payload;
         data.size = 1;
      }
      else
      {
         data.data = tok;
         data.size = strlen(tok) + 1;
      }

      rc = dbp->put(dbp, NULL, &key, &data, DB_NOOVERWRITE);
      if (rc != 0)
         printf("db_put() failed: call=[%s], payload=[%s], error=%s\n",
                (char *)key.data, (char *)data.data, db_strerror(rc));  
      else
         calls_added++;
   }

   fclose(fp);
   printf("Closed input text file %s\n", argv[2]);

   if (dbp != NULL)
   {
      dbp->close(dbp, 0);
      printf("Closed database %s\n", argv[1]);

      printf("Finished processing input text file %s, %lu calls were added\n",
              argv[2], calls_added); 

      return 0;
   }
   printf("Invalid database handle...database %s not valid, do not use it\n",argv[1]);
   return 1; 
}


