
/* KI4LKF */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/inotify.h>

#define CMD_BUF_SIZE 4096
static int fd = -1;
static int wd = -1;
static char an_event[1024];
static char cmd_buf[CMD_BUF_SIZE];
static bool keep_running = true;
static char temp_file[1024];

static void traceit(const char *fmt,...)
{
   time_t ltime;
   struct tm mytm;
   const short BFSZ = 1024;
   char buf[BFSZ];

   time(&ltime);
   localtime_r(&ltime,&mytm);

   snprintf(buf,BFSZ - 1,"%02d%02d%02d at %02d:%02d:%02d:",
            mytm.tm_mon+1,mytm.tm_mday,mytm.tm_year % 100,
            mytm.tm_hour,mytm.tm_min,mytm.tm_sec);

   va_list args;
   va_start(args,fmt);
   vsnprintf(buf + strlen(buf), BFSZ - strlen(buf) - 1, fmt, args);
   va_end(args);

   fprintf(stdout, "%s", buf);
   return;
}

int main(int argc, char *argv[])
{
  ssize_t len, i = 0;
  struct inotify_event *ev = NULL;
  char RADIO_ID[21];
  char *p = NULL;
  int delay = 5;

  setvbuf(stdout, (char *)NULL, _IOLBF, 0);

  traceit("Parameters passed: \n");
  for (i = 1; i < argc; i++)
     traceit("   %s\n", argv[i]);
  
  if (argc != 9)
  {
     traceit("Missing command line parameters...\n");
     return 1;
  }

  delay = atoi(argv[8]);
  if (delay <= 0)
     delay = 5;

  fd = inotify_init();
  if (fd == -1)
  {
     perror("inotify_init");
     return 1;
  }

  wd = inotify_add_watch(fd, argv[7], 
                         IN_CREATE | IN_ONLYDIR | IN_DONT_FOLLOW);
  if (wd == -1)
  {
     perror("inotify_add_watch");
     return 1;
  }

  traceit("waiting for notification files in folder %s\n", argv[7]);
  while (keep_running)
  {
     /* 
        I dont want to use a poll/select here, and I hope that the event descriptor is in BLOCKing mode
        after the kernel created it. I have another version that uses select() but it is
        not required since Linux creates that event descriptor in BLOCKing mode
     */
     
     len = read(fd, an_event, sizeof(an_event));
     i = 0;
       
     while (i < len) 
     {
        ev = (inotify_event *)&an_event[i]; 
        if (ev)
        {
           if ((ev->mask & IN_CREATE) && (ev->wd == wd) && (ev->len > 2))
           {
              traceit("Detected new file=[%s], name len=%d\n", ev->name, ev->len);

              /* example: A_linked.dat_LINKED_TO_XRF010_B */
              /* example: A_already_linked.dat_LINKED_TO_XRF010_B */
              /* example: A_unlinked.dat */
              /* example: A_failed_linked.dat */
              /* example: A_already_unlinked.dat */

              p = strstr(ev->name, ".dat");
              if (p)
              {
                 memset(RADIO_ID, '\0', 21);
                 if (p[4] == '_') 
                 {
                    strncpy(RADIO_ID, p + 5, 20);
                    p[4] = '\0';
                 }
                 strcpy(temp_file, argv[4]);
                 strcat(temp_file, "/");
                 strcat(temp_file, ev->name + 2);
                 snprintf(cmd_buf, CMD_BUF_SIZE - 1, "%s/dvtool_send %s %s %s \"%s\" %s %c %s > /dev/null",
                          argv[1], argv[2], argv[3], temp_file, RADIO_ID, argv[5], ev->name[0], argv[6]); 
                 traceit("Executing:%s\n", cmd_buf);
                 sleep(delay);
                 system(cmd_buf);
                 if (RADIO_ID[0] != '\0')
                    p[4] = '_';
              } 
           }
           strcpy(temp_file, argv[7]);
           strcat(temp_file, "/");
           strcat(temp_file, ev->name);
           unlink(temp_file);
        }
        else
        if ((ev->mask & IN_IGNORED) ||
            (ev->mask & IN_UNMOUNT))
        {
           traceit("Directory is now invalid\n");
           keep_running = false;
        }
        i += sizeof(struct inotify_event) + ev->len;
     }
     // The sleep is not required here, but just in case, the event descriptor is in non-BLOCKing mode 
     sleep(1);
  }

  inotify_rm_watch(fd,wd);
  close(fd);

  traceit("dextra_detect_notification is exiting\n");
  return 0;
}
