
/* by KI4LKF */

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>

#include <arpa/inet.h>
#include <netdb.h>

#include <net/if.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>

#include <linux/types.h>
#include <linux/filter.h>

/* Required for Binary search trees using C++ STL */
#include <string>
#include <map>
#include <set>
#include <utility>
using namespace std;

/* DB interface */
#include <db.h>

#define VERSION "v2.93"
#define OWNER_SIZE 8
#define IP_SIZE 15
#define MAXHOSTNAMELEN 64
#define CALL_SIZE 8

/* configuration data */
/* Put that in a structure, later */
static char OWNER[OWNER_SIZE + 1];
static char ADMIN[CALL_SIZE + 1];
static bool LINK_BY_USERS = false;
static int GW_PORT = 40000;
static int CTRL_PORT = 20000;
static int G2_INTERNAL_PORT = 20000;
static char INTERNAL_IF[IF_NAMESIZE + 1];
static char EXTERNAL_IF[IF_NAMESIZE + 1];
static char LISTEN_IP[IP_SIZE + 1];
static int LISTEN_PORT = 30001;
static int COMMAND_PORT = 30003;
static unsigned int MAX_USERS = 20;
static bool QSO_DETAILS = false;
static bool ANNOUNCE_LINK = true;
static char ANNOUNCE_DIR_NOTIFY[FILENAME_MAX + 1];
static char STATUS_FILE[FILENAME_MAX + 1];
static char LINK_CODE='O';
static char RELINK_CODE='W';
static char UNLINK_CODE='X';
static char IS_IT_LINKED_CODE='Z';
static short TIMEOUT = 30;
static short int REMOTE_REPLY = 10;
static char LINK_ON_STARTUP[CALL_SIZE + 1];
static bool DIFF = true;
static uint32_t INTERNAL_DADDR = 0;
static char USERS[FILENAME_MAX + 1];

static char ANNOUNCE_NOTIFY_FILE[FILENAME_MAX + 1];

static struct sockaddr_in toDst4;

struct a_user
{
   /* callsign of the connected station */
   char call[CALL_SIZE + 1];

   /* if true, packets from this call are dropped */
   bool isMute;

   /* is this a dongle user */
   bool is_dongle;

   /* IP address and UDP port of connected station */
   /* For easy access to the connected station */
   struct sockaddr_in sin;

   /* if countdown expires, the connection is terminated */
   short countdown;

   /* time when the station was connected */
   time_t connect_time;
};
/* The map of connected users */
/* The KEY is the unique IP address string of ip[IP_SIZE + 1] */
/* The data is a pointer to struct a_user */
typedef map<string, struct a_user *> a_user_list_type;
static a_user_list_type a_user_list;

/* Each one of the modules of this local repeater */
struct a_link
{
   /* Connected to which reflector */
   char to_call[CALL_SIZE + 1];

   /* Connected to which reflector module */
   char to_mod;

   /* IP address or hostname of the remote reflector */
   char to_ip[MAXHOSTNAMELEN + 1];

   /* repeater module */
   char from_mod;

   /* time link was requested */
   time_t link_time;

   /* time link was confirmed */
   time_t confirm_time;

   /* 
      Is this link persistent? 
      Only if we made the link request using the RELINK_CODE 
   */
   bool persist;
};
/* 
   This holds information about each module of the local repeater
   Index 0 is our module A
   Index 1 is our module B
   Index 2 is our module C
*/
static struct a_link mod_links[3];

/*
   The set of blocked callsigns.
   The KEY is the unique blocked callsign.
   Blocked callsigns added by the administrator of the dextra_srv
*/
typedef set<string> blocks_type;
static blocks_type blocks;

/* 
   The set of admin callsigns that are allowed 
   to link a repeater module, using a radio 
*/
typedef set<string> admin_type;
static admin_type admin;
static bool admin_found = false;


/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

/* The set of IPaddr + streamid going from us to a linked system */
/* The KEY is the unique IPaddr + streamid */
typedef set<string> streamid_from_gwy_list_type;
/* one for each module */
static streamid_from_gwy_list_type streamid_from_gwy_list[3];
/* for testing streamid_from_gwy_list */
static char a_streamid_from_gwy_test[(sizeof(uint32_t) + 2 + 1)];

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

/* Input from connected stations */
static int srv_sock = -1;

/* repeater controller sock */
static int internal_raw_sock = -1;

/* g2 sock */
static int external_raw_sock = -1;

/* Input from administrator */
static int cmd_sock = -1;

static fd_set fdset;
static struct timeval tv;

static time_t tNow = 0;
/* one for each module */
static time_t last_active_to_usr_time[3] = { 0, 0, 0 };
static time_t HBinterStart = 0;
static time_t ConfirmInterStart = 0;

static unsigned char readBuffer[124];
static unsigned char readBuffer2[124];
static unsigned char readBuffer3[124];
static struct sockaddr_in fromUser;
static struct sockaddr_in fromCmd;
static struct sockaddr_ll sll;

static u_char *udp_payload;
static int udp_payload_len;

static u_char *udp_payload2;
static int udp_payload_len2;

static FILE *logfp = NULL;
static FILE *statusfp = NULL;

static unsigned char rawbytes[56];
static int rawbytes_len = 0;

char linked_remote_system[CALL_SIZE + 1];

/* DB globals */
static DB *dbp = NULL;
static DBT key, data;
static u_int32_t flags = 0;

static uint16_t net_CTRL_PORT = htons(20000);
static uint16_t net_G2_INTERNAL_PORT = htons(20000);
static uint16_t net_GW_PORT = htons(40000);

static bool keep_running = true;

static char rpt1_copy[CALL_SIZE];
static char rpt2_copy[CALL_SIZE];
static unsigned char PFCS_copy[2];

static unsigned char old_streamid[2] = { 0x00, 0x00 };

static unsigned short crc_tabccitt[256] =
{
   0x0000,0x1189,0x2312,0x329b,0x4624,0x57ad,0x6536,0x74bf,
   0x8c48,0x9dc1,0xaf5a,0xbed3,0xca6c,0xdbe5,0xe97e,0xf8f7,
   0x1081,0x0108,0x3393,0x221a,0x56a5,0x472c,0x75b7,0x643e,
   0x9cc9,0x8d40,0xbfdb,0xae52,0xdaed,0xcb64,0xf9ff,0xe876,
   0x2102,0x308b,0x0210,0x1399,0x6726,0x76af,0x4434,0x55bd,
   0xad4a,0xbcc3,0x8e58,0x9fd1,0xeb6e,0xfae7,0xc87c,0xd9f5,
   0x3183,0x200a,0x1291,0x0318,0x77a7,0x662e,0x54b5,0x453c,
   0xbdcb,0xac42,0x9ed9,0x8f50,0xfbef,0xea66,0xd8fd,0xc974,
   0x4204,0x538d,0x6116,0x709f,0x0420,0x15a9,0x2732,0x36bb,
   0xce4c,0xdfc5,0xed5e,0xfcd7,0x8868,0x99e1,0xab7a,0xbaf3,
   0x5285,0x430c,0x7197,0x601e,0x14a1,0x0528,0x37b3,0x263a,
   0xdecd,0xcf44,0xfddf,0xec56,0x98e9,0x8960,0xbbfb,0xaa72,
   0x6306,0x728f,0x4014,0x519d,0x2522,0x34ab,0x0630,0x17b9,
   0xef4e,0xfec7,0xcc5c,0xddd5,0xa96a,0xb8e3,0x8a78,0x9bf1,
   0x7387,0x620e,0x5095,0x411c,0x35a3,0x242a,0x16b1,0x0738,
   0xffcf,0xee46,0xdcdd,0xcd54,0xb9eb,0xa862,0x9af9,0x8b70,
   0x8408,0x9581,0xa71a,0xb693,0xc22c,0xd3a5,0xe13e,0xf0b7,
   0x0840,0x19c9,0x2b52,0x3adb,0x4e64,0x5fed,0x6d76,0x7cff,
   0x9489,0x8500,0xb79b,0xa612,0xd2ad,0xc324,0xf1bf,0xe036,
   0x18c1,0x0948,0x3bd3,0x2a5a,0x5ee5,0x4f6c,0x7df7,0x6c7e,
   0xa50a,0xb483,0x8618,0x9791,0xe32e,0xf2a7,0xc03c,0xd1b5,
   0x2942,0x38cb,0x0a50,0x1bd9,0x6f66,0x7eef,0x4c74,0x5dfd,
   0xb58b,0xa402,0x9699,0x8710,0xf3af,0xe226,0xd0bd,0xc134,
   0x39c3,0x284a,0x1ad1,0x0b58,0x7fe7,0x6e6e,0x5cf5,0x4d7c,
   0xc60c,0xd785,0xe51e,0xf497,0x8028,0x91a1,0xa33a,0xb2b3,
   0x4a44,0x5bcd,0x6956,0x78df,0x0c60,0x1de9,0x2f72,0x3efb,
   0xd68d,0xc704,0xf59f,0xe416,0x90a9,0x8120,0xb3bb,0xa232,
   0x5ac5,0x4b4c,0x79d7,0x685e,0x1ce1,0x0d68,0x3ff3,0x2e7a,
   0xe70e,0xf687,0xc41c,0xd595,0xa12a,0xb0a3,0x8238,0x93b1,
   0x6b46,0x7acf,0x4854,0x59dd,0x2d62,0x3ceb,0x0e70,0x1ff9,
   0xf78f,0xe606,0xd49d,0xc514,0xb1ab,0xa022,0x92b9,0x8330,
   0x7bc7,0x6a4e,0x58d5,0x495c,0x3de3,0x2c6a,0x1ef1,0x0f78
};

static void send_a_pkt_to_all(unsigned char *raw, int rawlen, uint32_t test_addr);
static void send_heartbeat();
static void check_confirm();
static void calcPFCS(unsigned char packet[56]);
static void prepare_rawbytes(u_char *pay, int paylen);
static bool get_ip(char *call, char *ip);
static bool link_command(short i, bool persist);
static bool unlink_command(short i);
static int srv_open();
static void traceit(const char *fmt,...);
static void handle_cmd(char *buf);
static void print_users();
static void print_version();
static void print_links_screen();
static void print_links_file();
static void mute_users(bool mute);
static bool mute_call(char *call, bool mute);
static void print_blocks();
static void print_admin();
static int read_config(char *);
static int internal_raw_open();
static int external_raw_open();
static int  cmd_open();
static void post_notification(char *fullpathname);
static void runit();
static void sigCatch(int signum);
static int open_users(char *filename);

/* Open DB */
static int open_users(char *filename)
{
   int rc = 0;
   rc = db_create(&dbp, NULL, 0);
   if (rc != 0)
   {
      traceit("db_create() failed: error=%s\n", db_strerror(rc));
      return rc;
   }

   flags = DB_RDONLY;
   rc = dbp->open(dbp,NULL,filename,NULL,DB_BTREE,flags,0);
   if (rc != 0)
   {
      traceit("db_open() failed: error=%s\n", db_strerror(rc));
      dbp->close(dbp, 0);
      dbp = NULL;
      return rc;
   }
   traceit("Database %s opened OK\n", filename);
   return rc;
}

/* 
   Create the empty filename that serves as a notification that
   some link or unlink activity occured
*/
static void post_notification(char *fullpathname)
{
   int fd = open(fullpathname, 
                 O_WRONLY | O_CREAT | O_EXCL, 
                 S_IRUSR | S_IWUSR);

   if (fd == -1)
   {
      if (errno != EEXIST)
         traceit("Failed to create notification file %s, error=%d, message=%s\n",
                  fullpathname, errno, strerror(errno));
   }
   else
      close(fd);

   return;
}

/* prepare to send the data to the Gateway */
static void prepare_rawbytes(u_char *pay, int paylen)
{
   if (paylen == 58)
   {
      rawbytes[0] = 0x44;
      rawbytes[1] = 0x53;
      rawbytes[2] = 0x56;
      rawbytes[3] = 0x54;

      rawbytes[4] = 0x10;
      rawbytes[5] = 0x00;
      rawbytes[6] = 0x00; 
      rawbytes[7] = 0x00;

      rawbytes[8]  = pay[10];
      rawbytes[9]  = pay[11]; // 0x00;
      rawbytes[10] = pay[12]; // 0x01;
      rawbytes[11] = pay[13]; // 0x00;

      /* 
      copy 2-byte streamid, control, 3-byte flag 
           rpt1, rpt2, ur, my, sfx
           and 2-byte PFCS
      */
      memcpy((char *)rawbytes + 12, pay + 14, 44);
      rawbytes_len = 56;
   }
   else
   if ((paylen == 29) || (paylen == 32))
   {
      rawbytes[4] = 0x20;
      rawbytes[8] = pay[10];
      rawbytes[9]  = pay[11];
      rawbytes[10] = pay[12];
      rawbytes[11] = pay[13];

      /* copy 2-byte streamid and control */
      memcpy((char *)rawbytes + 12, pay + 14, 3);

      /* copy audio and text */
      if (paylen == 29)
         memcpy((char *)rawbytes + 15, pay + 17, 12);
      else
         memcpy((char *)rawbytes + 15, pay + 20, 12);
      rawbytes_len = 27;
   }
}

/* compute checksum */
static void calcPFCS(unsigned char packet[56])
{
   unsigned short crc_dstar_ffff = 0xffff;
   unsigned short tmp, short_c;
   short int i;

   for (i = 15; i < 54 ; i++)
   {
      short_c = 0x00ff & (unsigned short)packet[i];
      tmp = (crc_dstar_ffff & 0x00ff) ^ short_c;
      crc_dstar_ffff = (crc_dstar_ffff >> 8) ^ crc_tabccitt[tmp];
   }
   crc_dstar_ffff =  ~crc_dstar_ffff;
   tmp = crc_dstar_ffff;

   packet[54] = (unsigned char)(crc_dstar_ffff & 0xff);
   packet[55] = (unsigned char)((tmp >> 8) & 0xff);

   return;
}

/* log the event */
static void traceit(const char *fmt,...)
{
   time_t ltime;
   struct tm tm;
   const short BFSZ = 256;
   char buf[BFSZ];

   time(&ltime);
   localtime_r(&ltime,&tm);

   snprintf(buf,BFSZ - 1,"%02d%02d%02d at %02d:%02d:%02d:",
            tm.tm_mon+1,tm.tm_mday,tm.tm_year % 100,
            tm.tm_hour,tm.tm_min,tm.tm_sec);

   va_list args;
   va_start(args,fmt);
   vsnprintf(buf + strlen(buf), BFSZ - strlen(buf) -1, fmt, args);
   va_end(args);

   fprintf(logfp, "%s", buf);
   return;
}

/* process configuration file */
static int read_config(char *cfgFile)
{
   short int valid_params = 24;
   short int params = 0;

   FILE *cnf = NULL;
   char inbuf[1024];
   char *p = NULL;
   admin_type::iterator pos;
   char *ptr;
   unsigned short int i;

   cnf = fopen(cfgFile, "r");
   if (!cnf)
   {
      traceit("Failed to open file %s\n", cfgFile);
      return 1;
   }

   traceit("Reading file %s\n", cfgFile);
   while (fgets(inbuf, 1020, cnf) != NULL)
   {
      if (strchr(inbuf, '#'))
         continue;

      p = strchr(inbuf, '\r');
      if (p)
         *p = '\0';
      p = strchr(inbuf, '\n');
      if (p)
         *p = '\0';

      p = strchr(inbuf, '=');
      if (!p)
         continue;
      *p = '\0';

      if (strcmp(inbuf,"OWNER") == 0)
      {
          memset(OWNER,' ', sizeof(OWNER));
          OWNER[OWNER_SIZE] = '\0';

          ptr = strchr(p + 1, ' ');
          if (ptr)
             *ptr = '\0';
          
          if ((strlen(p + 1) < 1) || (strlen(p + 1) > (OWNER_SIZE - 2)))
             traceit("OWNER value [%s] invalid\n", p + 1);
          else
          {
             memcpy(OWNER, p + 1, strlen(p + 1));

             for (i = 0; i < strlen(OWNER); i++)
                OWNER[i] = toupper(OWNER[i]);
 
             traceit("OWNER=[%s]\n",OWNER);
             params ++;
          }
      }
      else
      if (strcmp(inbuf,"INTERNAL_IF") == 0)
      {
         ptr = strchr(p + 1, ' ');
         if (ptr)
            *ptr = '\0';

         if (strlen(p + 1) < 1)
            traceit("INTERNAL_IF value [%s] invalid\n", p + 1);
         else
         {
            memset(INTERNAL_IF, '\0', sizeof(INTERNAL_IF));
            strncpy(INTERNAL_IF, p + 1,IF_NAMESIZE);
            traceit("INTERNAL_IF=[%s]\n",INTERNAL_IF);
            params ++;
         }
      }
      else
      if (strcmp(inbuf,"EXTERNAL_IF") == 0)
      {
         ptr = strchr(p + 1, ' ');
         if (ptr)
            *ptr = '\0';

         if (strlen(p + 1) < 1)
            traceit("EXTERNAL_IF value [%s] invalid\n", p + 1);
         else
         {
            memset(EXTERNAL_IF, '\0', sizeof(EXTERNAL_IF));
            strncpy(EXTERNAL_IF, p + 1,IF_NAMESIZE);
            traceit("EXTERNAL_IF=[%s]\n",EXTERNAL_IF);
            params ++;
         }
      }
      else
      if (strcmp(inbuf,"LISTEN_IP") == 0)
      {
         ptr = strchr(p + 1, ' ');
         if (ptr)
            *ptr = '\0';

         if (strlen(p + 1) < 1)
            traceit("LISTEN_IP value [%s] invalid\n", p + 1);
         else
         {
            memset(LISTEN_IP, '\0', sizeof(LISTEN_IP));
            strncpy(LISTEN_IP, p + 1, IP_SIZE);
            traceit("LISTEN_IP=[%s]\n",LISTEN_IP);
            params ++;
         }
      }
      else
      if (strcmp(inbuf,"LISTEN_PORT") == 0)
      {
         LISTEN_PORT = atoi(p + 1);
         traceit("LISTEN_PORT=[%d]\n",LISTEN_PORT);
         params ++;
      }
      else
      if (strcmp(inbuf,"COMMAND_PORT") == 0)
      {
         COMMAND_PORT = atoi(p + 1);
         traceit("COMMAND_PORT=[%d]\n",COMMAND_PORT);
         params ++;
      }
      else
      if (strcmp(inbuf,"GW_PORT") == 0)
      {
         GW_PORT = atoi(p + 1);
         traceit("GW_PORT=[%d]\n",GW_PORT);
         params ++;
      }
      else
      if (strcmp(inbuf,"CTRL_PORT") == 0)
      {
         CTRL_PORT = atoi(p + 1);
         traceit("CTRL_PORT=[%d]\n",CTRL_PORT);
         params ++;
      }
      else
      if (strcmp(inbuf,"G2_INTERNAL_PORT") == 0)
      {
         G2_INTERNAL_PORT = atoi(p + 1);
         traceit("G2_INTERNAL_PORT=[%d]\n",G2_INTERNAL_PORT);
         params ++;
      }
      else
      if (strcmp(inbuf,"MAX_USERS") == 0)
      {
         MAX_USERS = atoi(p + 1);
         traceit("MAX_USERS=[%d]\n",MAX_USERS);
         params ++;
      }
      else
      if (strcmp(inbuf,"USERS") == 0)
      {
         memset(USERS,  '\0', sizeof(USERS));
         strncpy(USERS, p + 1,FILENAME_MAX);
         traceit("USERS=[%s]\n", USERS);
         params ++;
      }
      else
      if (strcmp(inbuf,"STATUS_FILE") == 0)
      {
         memset(STATUS_FILE, '\0', sizeof(STATUS_FILE));
         strncpy(STATUS_FILE, p + 1,FILENAME_MAX);
         traceit("STATUS_FILE=[%s]\n",STATUS_FILE);
         params ++;
      }
      else
      if (strcmp(inbuf,"QSO_DETAILS") == 0)
      {
         if (*(p + 1) == 'Y')
            QSO_DETAILS = true;
         else
            QSO_DETAILS = false;
         traceit("QSO_DETAILS=[%c]\n", *(p + 1));
         params ++;
      }
      else
      if (strcmp(inbuf,"LINK_BY_USERS") == 0)
      {
         if (*(p + 1) == 'Y')
            LINK_BY_USERS = true;
         else
            LINK_BY_USERS = false;
         traceit("LINK_BY_USERS=[%c]\n", *(p + 1));
         params ++;
      }
      else
      if (strcmp(inbuf,"ANNOUNCE_LINK") == 0)
      {
         if (*(p + 1) == 'Y')
            ANNOUNCE_LINK = true;
         else
            ANNOUNCE_LINK = false;
         traceit("ANNOUNCE_LINK=[%c]\n", *(p + 1));
         params ++;
      }
      else
      if (strcmp(inbuf,"ANNOUNCE_DIR_NOTIFY") == 0)
      {
         memset(ANNOUNCE_DIR_NOTIFY, '\0', sizeof(ANNOUNCE_DIR_NOTIFY));
         strncpy(ANNOUNCE_DIR_NOTIFY, p + 1, FILENAME_MAX);
         traceit("ANNOUNCE_DIR_NOTIFY=[%s]\n", ANNOUNCE_DIR_NOTIFY);
         params ++;
      }
      else
      if (strcmp(inbuf,"LINK_CODE") == 0)
      {
         LINK_CODE=toupper(*(p + 1));
         traceit("LINK_CODE=[%c]\n", LINK_CODE);
         params ++;
      }
      else
      if (strcmp(inbuf,"RELINK_CODE") == 0)
      {
         RELINK_CODE=toupper(*(p + 1));
         traceit("RELINK_CODE=[%c]\n", RELINK_CODE);
         params ++;
      }
      else
      if (strcmp(inbuf,"UNLINK_CODE") == 0)
      {
         UNLINK_CODE=toupper(*(p + 1));
         traceit("UNLINK_CODE=[%c]\n", UNLINK_CODE);
         params ++;
      }
      else
      if (strcmp(inbuf,"IS_IT_LINKED_CODE") == 0)
      {
         IS_IT_LINKED_CODE=toupper(*(p + 1));
         traceit("IS_IT_LINKED_CODE=[%c]\n", IS_IT_LINKED_CODE);
         params ++;
      }
      else
      if (strcmp(inbuf,"REMOTE_REPLY") == 0)
      {
         REMOTE_REPLY=atoi(p + 1);
         if (REMOTE_REPLY < 10)
         {
            traceit("Setting REMOTE_REPLY to 10\n");
            REMOTE_REPLY = 10;
         }
         traceit("REMOTE_REPLY=[%d]\n", REMOTE_REPLY);
         params ++;
      }
      else
      if (strcmp(inbuf,"LINK_ON_STARTUP") == 0)
      {
         memset(LINK_ON_STARTUP, '\0', sizeof(LINK_ON_STARTUP));
         strncpy(LINK_ON_STARTUP, p + 1, CALL_SIZE);
         traceit("LINK_ON_STARTUP=[%s]\n", LINK_ON_STARTUP);
         params ++;
      }
      else
      if (strcmp(inbuf,"DIFF") == 0)
      {
         if (*(p + 1) == 'Y')
            DIFF = true;
         else
            DIFF = false;
         traceit("DIFF=[%c]\n", *(p + 1));
         params ++;
      }
      else
      if (strcmp(inbuf,"ADMIN") == 0)
      {
         if (!admin_found)
         {
            admin_found = true;
            params ++;
         }
     
         memset(ADMIN,' ', CALL_SIZE);
         ADMIN[CALL_SIZE] = '\0';

         if ( (strlen(p + 1) < 1) || (strlen(p + 1) > CALL_SIZE) )
            traceit("ADMIN value [%s] invalid\n", p + 1);
         else
         {
            memcpy(ADMIN, p + 1, strlen(p + 1));

            for (i = 0; i < strlen(ADMIN); i++)
               ADMIN[i] = toupper(ADMIN[i]);

            traceit("ADMIN=[%s]\n",ADMIN);

            /* check for duplicates */
            pos = admin.find(ADMIN);
            if (pos != admin.end()) 
               traceit("[%s] already an administrator\n", ADMIN);
            else
            {
               if (admin.insert(ADMIN).second)
                  traceit("[%s] is now an administrator\n", ADMIN);
               else
                  traceit("failed to add [%s] as an administrator\n", ADMIN);
            }
         } 
      }
   }
   fclose(cnf);

   if (params != valid_params)
   {
      traceit("Configuration file %s invalid\n",cfgFile);
      return 1;
   } 

   if (LINK_CODE == UNLINK_CODE)
   {
      traceit("Error: LINK_CODE must be different from UNLINK_CODE\n");
      return 1;
   }

   if (LINK_CODE == RELINK_CODE)
   {
      traceit("Error: LINK_CODE must be different from RELINK_CODE\n");
      return 1;
   }

   if (LINK_CODE == IS_IT_LINKED_CODE)
   {
      traceit("Error: LINK_CODE must be different from IS_IT_LINKED_CODE\n");
      return 1;
   }

   if (UNLINK_CODE == RELINK_CODE)
   {
      traceit("Error: UNLINK_CODE must be different from RELINK_CODE\n");
      return 1;
   }

   if (UNLINK_CODE == IS_IT_LINKED_CODE)
   {
      traceit("Error: UNLINK_CODE must be different from IS_IT_LINKED_CODE\n");
      return 1;
   }

   if (RELINK_CODE == IS_IT_LINKED_CODE)
   {
      traceit("Error: RELINK_CODE must be different from IS_IT_LINKED_CODE\n");
      return 1;
   }

   if (GW_PORT == CTRL_PORT)
   {
      traceit("Error: GW_PORT must be different from CTRL_PORT\n");
      return 1;
   }

   if (GW_PORT == COMMAND_PORT)
   {
      traceit("Error: GW_PORT must be different from COMMAND_PORT\n");
      return 1;
   }

   if (GW_PORT == LISTEN_PORT)
   {
      traceit("Error: GW_PORT must be different from LISTEN_PORT\n");
      return 1;
   }

   if (CTRL_PORT == COMMAND_PORT)
   {
      traceit("Error: CTRL_PORT must be different from COMMAND_PORT\n");
      return 1;
   }

   if (CTRL_PORT == LISTEN_PORT)
   {
      traceit("Error: CTRL_PORT must be different from LISTEN_PORT\n");
      return 1;
   }

   if (COMMAND_PORT == LISTEN_PORT)
   {
      traceit("Error: COMMAND_PORT must be different from LISTEN_PORT\n");
      return 1;
   }

   net_CTRL_PORT = htons(CTRL_PORT);
   net_G2_INTERNAL_PORT = htons(G2_INTERNAL_PORT);
   net_GW_PORT = htons(GW_PORT);

   return 0;
}

/* check for any outstanding link requests */
static void check_confirm()
{
   short int i;
   time_t t_now;
   bool result = false;
   char from_mod = ' ';
 
   for (i = 0; i < 3; i++)
   {
      /* entry has a link request? */
      if (mod_links[i].to_mod != '\0')
      {
         /* check if it has been confirmed */
         if (mod_links[i].confirm_time < mod_links[i].link_time)
         {
            /* link request not confirmed yet */
            time(&t_now);

            /* Is it more than REMOTE_REPLY seconds since link was requested? */
            if ((t_now - mod_links[i].link_time) > REMOTE_REPLY)
            {
               /* 
                  Assume remote system is dead or it will not honor our link request
                  or its IP address has changed or ...
               */

               traceit("Module %c un-linked from %s %c %s due to timeout on link request\n",
                       mod_links[i].from_mod, mod_links[i].to_call,
                       mod_links[i].to_mod, mod_links[i].to_ip);

               if (mod_links[i].persist)
               {
                  from_mod = mod_links[i].from_mod;
                  result = link_command(i, true);
                  if (!result)
                  {
                     if (ANNOUNCE_LINK)
                     {
                        snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_failed_linked.dat",
                                 ANNOUNCE_DIR_NOTIFY, from_mod);
                        post_notification(ANNOUNCE_NOTIFY_FILE);
                     }
                     traceit("Failed to send link command\n");
                  }
               }
               else
               {
                  if (ANNOUNCE_LINK)
                  {
                     snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_unlinked.dat",
                              ANNOUNCE_DIR_NOTIFY, mod_links[i].from_mod);
                     post_notification(ANNOUNCE_NOTIFY_FILE);
                  }
                  mod_links[i].from_mod = '\0';
                  mod_links[i].to_mod = '\0';
                  mod_links[i].to_ip[0] = '\0';
                  mod_links[i].to_call[0] = '\0';
                  mod_links[i].link_time = 0;
                  mod_links[i].confirm_time = 0; 
                  mod_links[i].persist = false;
               }

               print_links_file();
            } 
         }
      }
   }
}

/* send keepalive packet */
static void send_heartbeat()
{
   a_user_list_type::iterator pos;
   struct a_user *a_user_ptr;
   short i;
   struct a_link temp_mod_links[3];
   bool a_link_was_dropped = false;
   char from_mod = ' ';
   bool result = false;

   for (i = 0; i < 3; i++)
      memset(&temp_mod_links[i], '\0', sizeof(temp_mod_links[i]));

   for (pos = a_user_list.begin(); pos != a_user_list.end(); pos++)
   {
      a_user_ptr = (struct a_user *)pos->second;

      sendto(srv_sock, OWNER, strlen(OWNER) + 1,
             0,(struct sockaddr *)&(a_user_ptr->sin),
             sizeof(struct sockaddr_in));

      if (a_user_ptr->countdown >= 0)
         a_user_ptr->countdown --;

      if (a_user_ptr->countdown < 0)
      {
         traceit("call=%s timeout, removing %s, users=%d\n",
                 a_user_ptr->call,
                 pos->first.c_str(),
                 a_user_list.size() - 1);

         /* remove any links if any of our repeater modules are linked to it */
         for (i = 0; i < 3; i++)
         {
            if (strcmp(mod_links[i].to_ip, pos->first.c_str()) == 0)
            {
               a_link_was_dropped = true;
               if (mod_links[i].persist)
               {
                  temp_mod_links[i].from_mod = mod_links[i].from_mod;
                  temp_mod_links[i].to_mod = mod_links[i].to_mod;
                  strcpy(temp_mod_links[i].to_call, mod_links[i].to_call);
                  strcpy(temp_mod_links[i].to_ip, mod_links[i].to_ip);
                  temp_mod_links[i].link_time = mod_links[i].link_time;
                  temp_mod_links[i].confirm_time = mod_links[i].confirm_time;
                  temp_mod_links[i].persist = mod_links[i].persist;
               }
               else
               {
                  if (ANNOUNCE_LINK)
                  {
                     snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_unlinked.dat",
                              ANNOUNCE_DIR_NOTIFY, mod_links[i].from_mod);
                     post_notification(ANNOUNCE_NOTIFY_FILE);
                  }
               }
               mod_links[i].from_mod = '\0';
               mod_links[i].to_mod = '\0';
               mod_links[i].to_call[0] = '\0';
               mod_links[i].to_ip[0] = '\0';
               mod_links[i].link_time = 0;
               mod_links[i].confirm_time = 0;
               mod_links[i].persist = false;
            }
         }

         /* Remove the user */
         free(pos->second);
         pos->second = NULL;
         a_user_list.erase(pos);

      }
   }

   if (a_link_was_dropped)
   {
      for (i = 0; i < 3; i++)
      {
         if (temp_mod_links[i].to_mod != '\0')
         {
            if (temp_mod_links[i].persist)
            {
               mod_links[i].from_mod = temp_mod_links[i].from_mod;
               mod_links[i].to_mod = temp_mod_links[i].to_mod;
               strcpy(mod_links[i].to_call, temp_mod_links[i].to_call);
               strcpy(mod_links[i].to_ip, temp_mod_links[i].to_ip);
               mod_links[i].link_time = temp_mod_links[i].link_time;
               mod_links[i].confirm_time = temp_mod_links[i].confirm_time;
               mod_links[i].persist = temp_mod_links[i].persist;

               from_mod = mod_links[i].from_mod;
               result = link_command(i, true);
               if (!result)
               {
                  if (ANNOUNCE_LINK)
                  {
                     snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_failed_linked.dat",
                              ANNOUNCE_DIR_NOTIFY, from_mod);
                     post_notification(ANNOUNCE_NOTIFY_FILE);
                  }
                  traceit("Failed to send link command\n");
               }
            }
         }
      }
      print_links_file();
   }

   return;
}

/* relay data to all stations */
static void send_a_pkt_to_all(unsigned char *raw, int rawlen, uint32_t test_addr)
{
   short i,j;
   bool is_pass = true;
   streamid_from_gwy_list_type::iterator from_gwy_pos;
   a_user_list_type::iterator user_pos;
   struct a_user *a_user_ptr;

   if (rawlen == 56)
   {
      /* save rpt1, rpt2, PFCS */
      memcpy(rpt1_copy, (char *)raw + 18, CALL_SIZE);
      memcpy(rpt2_copy, (char *)raw + 26, CALL_SIZE);
      memcpy(PFCS_copy, raw + 54, 2);

      if ((memcmp(raw + 42, OWNER, 7) != 0) &&         /* block repeater announcements   */

          /*** CQ calls only, no routed calls ***/
          (memcmp(raw + 34, "CQCQCQ", 6) == 0) &&

          ((raw[15] == 0x00) ||    /* normal */
           (raw[15] == 0x08) ||    /* EMR */
           (raw[15] == 0x20) ||    /* BK */
           (raw[15] == 0x28)) &&   /* EMR + BK */

          (memcmp(raw + 18, OWNER, 7) == 0) &&         /* rpt1 must be OWNER */
          (memcmp(raw + 26, OWNER, 7) == 0) &&         /* rpt2 must be OWNER */
          (raw[33] == 'G'))
         is_pass = true;
      else
         is_pass = false;
   }

   for (user_pos = a_user_list.begin(); user_pos != a_user_list.end(); user_pos++)
   {
      a_user_ptr = (struct a_user *)user_pos->second;
      if (test_addr != a_user_ptr->sin.sin_addr.s_addr)
      {
         if (rawlen == 56)
         {
            if (is_pass)
            {
               if (a_user_ptr->is_dongle)
                  sendto(srv_sock, (char *)raw,rawlen,
                         0,(struct sockaddr *)&(a_user_ptr->sin),
                         sizeof(struct sockaddr_in));
               else
               {
                  /* Remote station is a linked system */
                  if (raw[25] == 'A')
                     i = 0;
                  else
                  if (raw[25] == 'B')
                     i = 1;
                  else
                  if (raw[25] == 'C')
                     i = 2;
                  else
                     i = -1;

                  if (i >= 0)
                  {
                     if ((mod_links[i].to_mod != '\0') &&
                         (strcmp(mod_links[i].to_call, a_user_ptr->call) == 0))
                     {
                        memcpy((char *)raw + 18, mod_links[i].to_call, CALL_SIZE);
                        raw[25] = mod_links[i].to_mod;
                        memcpy((char *)raw + 26, mod_links[i].to_call, CALL_SIZE);
                        raw[33] = 'G';
                        calcPFCS(raw);

                        memcpy(a_streamid_from_gwy_test,
                            (unsigned char *)&(a_user_ptr->sin.sin_addr.s_addr),
                            sizeof(uint32_t));
                        memcpy(a_streamid_from_gwy_test + sizeof(uint32_t),raw + 12, 2);
                        a_streamid_from_gwy_test[sizeof(uint32_t) + 2] = '\0'; 
 
                        from_gwy_pos = streamid_from_gwy_list[i].find(a_streamid_from_gwy_test);
                        if (from_gwy_pos == streamid_from_gwy_list[i].end())
                        {
                           if (streamid_from_gwy_list[i].insert(a_streamid_from_gwy_test).second)
                           {
                              /* send header 5 times, because UDP packets can get lost */
                              for (j = 0; j < 5; j++)
                                 sendto(srv_sock, (char *)raw,rawlen,
                                        0,(struct sockaddr *)&(a_user_ptr->sin),
                                        sizeof(struct sockaddr_in));

                              time(&last_active_to_usr_time[i]);
                           }
                           else
                              traceit("Failed to add %s and streamids %d,%d\n",
                                      mod_links[i].to_ip, raw[12], raw[13]);
                        }
                        else
                           time(&last_active_to_usr_time[i]);

                        /* back to original values */
                        memcpy((char *)raw + 18, rpt1_copy, CALL_SIZE);
                        memcpy((char *)raw + 26, rpt2_copy, CALL_SIZE);
                        memcpy(raw + 54, PFCS_copy, 2);
                     }
                  }
               }
            }
         }
         else
         if (rawlen == 27)
         {
            if (a_user_ptr->is_dongle)
               sendto(srv_sock, (char *)raw,rawlen,
                      0,(struct sockaddr *)&(a_user_ptr->sin),
                      sizeof(struct sockaddr_in));
            else
            {
               memcpy(a_streamid_from_gwy_test,
                      (unsigned char *)&(a_user_ptr->sin.sin_addr.s_addr),
                      sizeof(uint32_t));
               memcpy(a_streamid_from_gwy_test + sizeof(uint32_t), raw + 12, 2);
               a_streamid_from_gwy_test[sizeof(uint32_t) + 2] = '\0';

               for (i = 0; i < 3; i++)
               {
                  from_gwy_pos = streamid_from_gwy_list[i].find(a_streamid_from_gwy_test);
                  if (from_gwy_pos != streamid_from_gwy_list[i].end())
                  {
                     sendto(srv_sock, (char *)raw,rawlen,
                            0,(struct sockaddr *)&(a_user_ptr->sin),
                            sizeof(struct sockaddr_in));

                     if ((raw[14] & 0x40) != 0)
                        streamid_from_gwy_list[i].erase(from_gwy_pos);

                     time(&last_active_to_usr_time[i]);

                     break;
                  }
               }
            }
         }
      }
   }
   return;
}

/* Open g2 */
static int external_raw_open()
{
   struct ifreq ifr;
   int reuse = 1;

   struct sockaddr_in *paddr;

   /*** packet filter: tcpdump -dd udp dst port 40000 ***/
   /*** dest is our local G2 udp port 40000 ***/
   struct sock_filter BPF_code[]=
   {
	{ 0x28, 0, 0, 0x0000000c },
	{ 0x15, 0, 4, 0x000086dd },
	{ 0x30, 0, 0, 0x00000014 },
	{ 0x15, 0, 11, 0x00000011 },
	{ 0x28, 0, 0, 0x00000038 },
	{ 0x15, 8, 9, 0x00009c40 },
	{ 0x15, 0, 8, 0x00000800 },
	{ 0x30, 0, 0, 0x00000017 },
	{ 0x15, 0, 6, 0x00000011 },
	{ 0x28, 0, 0, 0x00000014 },
	{ 0x45, 4, 0, 0x00001fff },
	{ 0xb1, 0, 0, 0x0000000e },
	{ 0x48, 0, 0, 0x00000010 },
	{ 0x15, 0, 1, 0x00009c40 },
	{ 0x6, 0, 0, 0x00000060 },
	{ 0x6, 0, 0, 0x00000000 }
   };
   struct sock_fprog Filter;
   Filter.filter = BPF_code;
   Filter.len = 16;

   external_raw_sock = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_IP));
   if (external_raw_sock == -1)
   {
      traceit("Failed to create external raw socket, errno=%d\n", errno);
      return 1;
   }

   /***
   if (setsockopt(external_raw_sock, SOL_SOCKET, SO_ATTACH_FILTER, &Filter, sizeof(Filter)) <0)
   {
      traceit("Failed to attach packet filter for external raw socket, errno=%d\n", errno);
      close(external_raw_sock);
      return 1;
   }
   ***/

   if (setsockopt(external_raw_sock,SOL_SOCKET,SO_REUSEADDR, (char *)&reuse, sizeof(reuse)) == -1)
   {
      traceit("setsockopt REUSE external raw socket failed,errno=%d\n",errno);
      close(external_raw_sock);
      return 1;
   }

   memset(&ifr, '\0', sizeof(ifr));
   strncpy((char *)ifr.ifr_name, EXTERNAL_IF, IFNAMSIZ);
   if ((ioctl(external_raw_sock, SIOCGIFADDR, &ifr)) == -1)
   {
      traceit("SIOCGIFADDR failed on EXTERNAL_IF %s, errno=%d\n",EXTERNAL_IF,errno);
      close(external_raw_sock);
      return 1;
   }

   paddr = (struct sockaddr_in *)&(ifr.ifr_addr);
   traceit("EXTERNAL_IF %s has private IP %s\n",
           EXTERNAL_IF, inet_ntoa(paddr->sin_addr));

   memset(&toDst4,0,sizeof(struct sockaddr_in));
   toDst4.sin_family = AF_INET;
   toDst4.sin_port = htons(GW_PORT);
   toDst4.sin_addr.s_addr = paddr->sin_addr.s_addr;

   memset(&sll, '\0', sizeof(sll));
   sll.sll_family = AF_PACKET;
   sll.sll_ifindex =  if_nametoindex(EXTERNAL_IF);
   sll.sll_protocol = htons(ETH_P_ALL);
   if ((bind(external_raw_sock, (struct sockaddr *)&sll, sizeof(sll)))== -1)
   {
      traceit("Failed to bind external raw socket on %s, errno=%d\n",EXTERNAL_IF,errno);
      close(external_raw_sock);
      return 1;
   }

   fcntl(external_raw_sock,F_SETFL,O_NONBLOCK);
   return 0;
}


/* Open repeater-controler */
static int internal_raw_open()
{
   struct ifreq ifr;
   int reuse = 1;
   struct sockaddr_in *paddr; 

   internal_raw_sock = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_IP));
   if (internal_raw_sock == -1)
   {
      traceit("Failed to create internal raw socket, errno=%d\n", errno);
      return 1;
   }

   if (setsockopt(internal_raw_sock,SOL_SOCKET,SO_REUSEADDR, (char *)&reuse, sizeof(reuse)) == -1)
   {
      traceit("setsockopt REUSE internal raw socket failed,errno=%d\n",errno);
      close(internal_raw_sock);
      return 1;
   } 

   memset(&ifr, '\0', sizeof(ifr));
   strncpy((char *)ifr.ifr_name, INTERNAL_IF, IFNAMSIZ);
   if ((ioctl(internal_raw_sock, SIOCGIFADDR, &ifr)) == -1)
   {
      traceit("SIOCGIFADDR failed on INTERNAL_IF %s, errno=%d\n",INTERNAL_IF,errno);
      close(internal_raw_sock);
      return 1;
   }

   paddr = (struct sockaddr_in*)&(ifr.ifr_addr);
   traceit("INTERNAL_IF %s has private IP %s\n", 
           INTERNAL_IF, inet_ntoa(paddr->sin_addr)); 

   INTERNAL_DADDR = paddr->sin_addr.s_addr;

   memset(&sll, '\0', sizeof(sll));
   sll.sll_family = AF_PACKET;
   
   if (DIFF)
      sll.sll_ifindex = if_nametoindex(INTERNAL_IF);
   else
      sll.sll_ifindex = if_nametoindex("lo");

   sll.sll_protocol = htons(ETH_P_ALL);
   if ((bind(internal_raw_sock, (struct sockaddr *)&sll, sizeof(sll)))== -1)
   {
      traceit("Failed to bind internal raw socket on %s, errno=%d\n",
              (DIFF)?INTERNAL_IF:"lo",errno);
      close(internal_raw_sock);
      return 1;
   }

   fcntl(internal_raw_sock,F_SETFL,O_NONBLOCK);
   return 0;
}

/* Open socket for input from stations */
static int srv_open()
{
   struct sockaddr_in sin;

   srv_sock = socket(PF_INET,SOCK_DGRAM,0);
   if (srv_sock == -1)
   {
      traceit("Failed to create server socket,errno=%d\n",errno);
      return 1;
   }
   memset(&sin,0,sizeof(struct sockaddr_in));
   sin.sin_family = AF_INET;
   sin.sin_port = htons(LISTEN_PORT);
   sin.sin_addr.s_addr = inet_addr(LISTEN_IP);

   if (bind(srv_sock,(struct sockaddr *)&sin,sizeof(struct sockaddr_in)) != 0)
   {
      traceit("Failed to bind server socket on port %s,%d, errno=%d\n",
              LISTEN_IP, LISTEN_PORT,errno);
      close(srv_sock);
      return 1;
   }
   fcntl(srv_sock,F_SETFL,O_NONBLOCK);
   return 0;
}

/* Open the COMMAND port */
static int cmd_open()
{
   struct sockaddr_in sin;

   cmd_sock = socket(PF_INET,SOCK_DGRAM,0);
   if (cmd_sock == -1)
   {
      traceit("Failed to create cmd socket,errno=%d\n",errno);
      return 1;
   }
   memset(&sin,0,sizeof(struct sockaddr_in));
   sin.sin_family = AF_INET;
   sin.sin_port = htons(COMMAND_PORT);
   sin.sin_addr.s_addr = inet_addr(LISTEN_IP);

   if (bind(cmd_sock,(struct sockaddr *)&sin,sizeof(struct sockaddr_in)) != 0)
   {
      traceit("Failed to bind cmd socket on IP %s, port %d, errno=%d\n",
              LISTEN_IP, COMMAND_PORT,errno);
      close(cmd_sock);
      return 1;
   }
   fcntl(cmd_sock,F_SETFL,O_NONBLOCK);
   return 0;
}

/* send a request to link to a reflector module */
static bool link_command(short i, bool persist)
{
   short counter;
   char link_request[CALL_SIZE + 3];
   struct sockaddr_in sin;

   if ((i < 0) || (i > 2))
   {
      traceit("Invalid index %d for mod_links\n",i);
      return false;
   }

   /* check if another one of our repeater modules is already linked to the same reflector module */
   for (counter = 0; counter < 3; counter++)
   {
      if (counter != i)
      {
         if ((mod_links[counter].to_mod == mod_links[i].to_mod) &&
             (strcmp(mod_links[counter].to_call, mod_links[i].to_call) == 0))
         {
            traceit("Another mod %c is already linked to %s %c\n",
                     mod_links[counter].from_mod, 
                     mod_links[counter].to_call,
                     mod_links[counter].to_mod); 
            break;
         }
      }
   }
   if (counter < 3)
   {
      mod_links[i].to_ip[0] = '\0';
      mod_links[i].to_call[0] = '\0';
      mod_links[i].from_mod = '\0';
      mod_links[i].to_mod = '\0';
      mod_links[i].link_time = 0;
      mod_links[i].confirm_time = 0;
      mod_links[i].persist = false;
      
      return false;
   }

   if (memcmp(mod_links[i].to_call, OWNER, 8) == 0)
   {
      traceit("lnking to your own repeater %s is not allowed\n", mod_links[i].to_call);
        
      mod_links[i].to_ip[0] = '\0';
      mod_links[i].to_call[0] = '\0';
      mod_links[i].from_mod = '\0';
      mod_links[i].to_mod = '\0';
      mod_links[i].link_time = 0;
      mod_links[i].confirm_time = 0;
      mod_links[i].persist = false;
      
      return false;
   }

   mod_links[i].persist = persist;
   mod_links[i].to_ip[0] = '\0';

   /* get hostname/IP address of reflector */
   if (!get_ip(mod_links[i].to_call, mod_links[i].to_ip))
   {
      traceit("get_ip failed for call %s\n", mod_links[i].to_call);
    
      mod_links[i].to_ip[0] = '\0';
      mod_links[i].confirm_time = 0;
      mod_links[i].to_call[0] = '\0';
      mod_links[i].to_mod = '\0';
      mod_links[i].from_mod = '\0';
      mod_links[i].link_time = 0;
      mod_links[i].persist = false;
      
      return false;
   }
  
   /* Do we have the IP address for the reflector ? */ 
   if (mod_links[i].to_ip[0] == '\0')
   {
      traceit("call %s has no hostname\n",mod_links[i].to_call);
    
      mod_links[i].to_ip[0] = '\0';
      mod_links[i].confirm_time = 0;

      /* if the link was a persistent link, do not touch these */
      if (!mod_links[i].persist)
      {
         mod_links[i].to_call[0] = '\0';
         mod_links[i].to_mod = '\0';
         mod_links[i].from_mod = '\0';
         mod_links[i].link_time = 0;
         mod_links[i].persist = false;
      }
      return false;
   }

   traceit("Sending request from mod %c to link with %s %c, %s\n",
            mod_links[i].from_mod,mod_links[i].to_call, mod_links[i].to_mod,
            mod_links[i].to_ip);


   /* prepare to send the link request to the reflector */
   memset(&sin,0,sizeof(struct sockaddr_in));
   sin.sin_family = AF_INET;
   sin.sin_port = htons(LISTEN_PORT);
   sin.sin_addr.s_addr = inet_addr(mod_links[i].to_ip);

   strcpy(link_request, OWNER);
   link_request[8] = mod_links[i].from_mod;
   link_request[9] = mod_links[i].to_mod;
   link_request[10] = '\0';

   /* UDP is unreliable, send the link request a few times */
   for (counter = 0; counter < 5; counter++)
      sendto(srv_sock, link_request, CALL_SIZE + 3,
             0,(struct sockaddr *)&sin,
             sizeof(struct sockaddr_in));

   time(&(mod_links[i].link_time));
   mod_links[i].persist = persist;

   return true;
}

/* send a request to unlink from a reflector module */
static bool unlink_command(short i)
{
   short counter;
   char unlink_request[CALL_SIZE + 3];
   struct sockaddr_in sin;
   bool remove_user = false;

   char ip_to_delete[IP_SIZE + 1];
   char call_to_delete[CALL_SIZE + 1];

   if ((i < 0) || (i > 2))
   {
      traceit("Invalid index %d for mod_links\n",i);
      return false;
   }

   memset(&sin,0,sizeof(struct sockaddr_in));
   sin.sin_family = AF_INET;
   sin.sin_port = htons(LISTEN_PORT);
   sin.sin_addr.s_addr = inet_addr(mod_links[i].to_ip);

   strcpy(unlink_request, OWNER);
   unlink_request[8] = mod_links[i].from_mod;
   unlink_request[9] = ' ';
   unlink_request[10] = '\0';

   /* save the reflector module info that maybe deleted ? */
   strcpy(ip_to_delete,mod_links[i].to_ip);
   strcpy(call_to_delete,mod_links[i].to_call);

   /* UDP is unreliable, send the un-link request a few times */
   for (counter = 0; counter < 5; counter++)
      sendto(srv_sock, unlink_request, CALL_SIZE + 3,
             0,(struct sockaddr *)&sin,
             sizeof(struct sockaddr_in));

   traceit("Module %c un-linked from %s %c %s\n",
           mod_links[i].from_mod, mod_links[i].to_call, 
           mod_links[i].to_mod, mod_links[i].to_ip);

   /* reset link information for our repeater module */
   mod_links[i].from_mod = '\0';
   mod_links[i].to_mod = '\0';
   mod_links[i].to_ip[0] = '\0';
   mod_links[i].to_call[0] = '\0';
   mod_links[i].link_time = 0;
   mod_links[i].confirm_time = 0;
   mod_links[i].persist = false;

   /* assume that all links to that reflector are clear */
   remove_user = true;
   /*
      Find out if any of our repeater modules
      is still linked to that reflector
   */
   for (i = 0; i < 3; i++)
   {
      if (strcmp(ip_to_delete, mod_links[i].to_ip) == 0)
      {
         remove_user = false;
         break;
      }
   }
   if (remove_user)
   {
      if (a_user_list.erase(ip_to_delete) > 0)
      {
         traceit("All links to station %s %s are off, removed\n",
                 call_to_delete, ip_to_delete);
      }
      else
         traceit("Tried to remove station %s %s, but could not, station not in list...OK\n",
                  call_to_delete, ip_to_delete);
   }
   return true;
}

/* Look up IP address in database, using the callsign as the key */
static bool get_ip(char *call, char *ip)
{
   int rc;
   char payload[MAXHOSTNAMELEN + 1];

   struct hostent *hp;
   struct in_addr ia;

   memset(&key, 0, sizeof(DBT));
   memset(&data, 0, sizeof(DBT));

   key.data = call;
   key.size = CALL_SIZE + 1;

   data.data = payload;
   data.size = MAXHOSTNAMELEN + 1;

   rc = dbp->get(dbp, NULL, &key, &data, 0);
   if (rc != 0)
      return false;

   if ((*((char *)data.data) != '\0') && ip)
   {
      hp = gethostbyname((char *)data.data);
      if (!hp)
         traceit("Call %s is host %s but could not resolve to IP\n",
                 call, (char *)data.data);
      else
      {
         memcpy(&ia, hp->h_addr,sizeof(struct in_addr));
         strcpy(ip, inet_ntoa(ia));
      }
   }
   return true;
}

/* Process the shell command */
static void handle_cmd(char *buf)
{
   char *p = NULL;
   char call[CALL_SIZE + 1];
   char cmd[4];
   short i, j;
   char nak[5];

   char *ptr_cmd = NULL;
   char *ptr_call = NULL;
   const char *delim = " ";

   blocks_type::iterator block_pos;
   admin_type::iterator admin_pos;
   bool result = false;

   int rc = -1;

   nak[0] = '\0';

   p = strchr(buf, '\r');
   if (p)
      *p = '\0';

   p = strchr(buf, '\n');
   if (p)
      *p = '\0';

   if (strlen(buf) < 2)
      return;

   traceit("Received command [%s] from [%s]\n", buf, inet_ntoa(fromCmd.sin_addr));

   if (strcmp(buf, "sh") == 0)  /* shut it down */
      keep_running = false;
   else
   if (strcmp(buf, "mu") == 0) /* mute users */
      mute_users(true);
   else
   if (strcmp(buf, "uu") == 0) /* un-mute users */
      mute_users(false);
   else
   if (strcmp(buf, "qsoy") == 0)  /* QSO_DETAILS yes */
      QSO_DETAILS = true;
   else
   if (strcmp(buf, "qson") == 0) /* QSO_DETAILS no */
      QSO_DETAILS = false;       
   else
   if (strcmp(buf, "al") == 0) /* allow links by users */
      LINK_BY_USERS = true;
   else
   if (strcmp(buf, "dl") == 0) /* disallow links by users */
      LINK_BY_USERS = false;
   else
   if (strcmp(buf, "pu") == 0)   /* print users */
      print_users();
   else
   if (strcmp(buf, "pv") == 0)   /* print version */
      print_version();
   else
   if (strcmp(buf, "pb") == 0)   /* print blocked users */
      print_blocks();
   else
   if (strcmp(buf, "pa") == 0)  /* print administrators */
      print_admin();
   else
   if (strcmp(buf, "pl") == 0)   /* print links */
      print_links_screen();
   else
   {
      ptr_cmd = strtok(buf, delim);
      if (ptr_cmd)
         ptr_call = strtok(NULL, delim);

      if (!ptr_cmd || !ptr_call)
      {
         traceit("Missing command parameters\n");
         strcpy(nak, "NAK\n");
      }
      else
      if (strcmp(ptr_cmd, "upd") == 0)
      {
         if (dbp != NULL)
         {
            dbp->close(dbp, 0);
            traceit("Closing database %s\n",USERS);
         }
         rc = rename(ptr_call, USERS);
         if (rc != 0)
            traceit("Failed to move %s to %s, error=%d\n", ptr_call, USERS, errno);
         else
            traceit("Moved %s over %s, OK\n", ptr_call, USERS); 

         rc = open_users(USERS);
         if (rc != 0)
         {
            traceit("Failed to re-open %s\n", USERS);
         
            /* Unlink all our modules */   
            for (i = 0; i < 3; i++)
            {
               if (mod_links[i].to_mod != '\0')
                  unlink_command(i);
            }
            print_links_file();
            traceit("dextra_srv exiting\n");
            fclose(logfp);
            exit(1);
         }
      }
      else
      if ((strlen(ptr_cmd) > 3) || (strlen(ptr_call) > CALL_SIZE))
      {
         traceit("Invalid command length\n");
         strcpy(nak, "NAK\n");
      }
      else
      {
         strcpy(cmd, ptr_cmd);
         strcpy(call, ptr_call);

         if (strlen(call) < CALL_SIZE)
         {
            for (i = strlen(call); i < CALL_SIZE; i++)
               call[i] = ' ';
            call[CALL_SIZE] = '\0';
         }

         if (strcmp(cmd, "ab") == 0)   /* Add block for a callsign, example: ab KI4LKF */
         {
            block_pos = blocks.find(call);
            if (block_pos != blocks.end())
            {
               traceit("Call %s already blocked\n", call);
               strcpy(nak, "NAK\n");
            }
            else
            {
               if (blocks.insert(call).second)
                  traceit("call %s is now blocked\n", call);
               else
               {
                  traceit("Failed to block call %s\n",call);
                  strcpy(nak, "NAK\n");
               }
            }
         }
         else
         if (strcmp(cmd, "rb") == 0)   /* remove block, example: rb KI4LKF */
         {
            if (blocks.erase(call) > 0)
               traceit("call %s is now unblocked\n", call);
            else
            {
               traceit("Could not find blocked call %s\n", call);
               strcpy(nak, "NAK\n");
            }
         }
         else
         if (strcmp(cmd, "aa") == 0)  /* add administrator, example: aa KI4LKF */
         {
            admin_pos = admin.find(call);
            if (admin_pos != admin.end())
            {
               traceit("Call %s already an admin\n", call);
               strcpy(nak, "NAK\n");
            }
            else
            {
               if (admin.insert(call).second)
                  traceit("call %s is now admin\n", call);
               else
               {
                  traceit("Failed to add call %s as an ADMIN\n",call);
                  strcpy(nak, "NAK\n");
               }
            }
         }
         else
         if (strcmp(cmd, "ra") == 0)   /* remove administrator, example: ra KI4LKF */
         {
            if (admin.erase(call) > 0)
               traceit("call %s removed from ADMIN list\n", call);
            else
            {
               traceit("Could not find ADMIN call %s\n", call);
               strcpy(nak, "NAK\n");
            }
         }
         else
         if (strcmp(cmd, "um") == 0)   /* un-link a repeater module, example: um B */
         {
            if (call[0] == 'A')
               i = 0;
            else
            if (call[0] == 'B')
              i = 1;
            else
            if (call[0] == 'C')
               i = 2;
            else
            {
               i = -1;
               traceit("The module %c is invalid\n", call[0]);
               strcpy(nak, "NAK\n");
            }
            if (i >= 0)
            {
               if (mod_links[i].to_mod == '\0')
               {
                  if (ANNOUNCE_LINK)
                  {
                     snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_already_unlinked.dat",
                              ANNOUNCE_DIR_NOTIFY,call[0]);
                     post_notification(ANNOUNCE_NOTIFY_FILE);
                  }
                  traceit("Module %c is already un-linked\n", call[0]);
                  strcpy(nak, "NAK\n");
               }
               else
               {
                  if (!unlink_command(i))
                  {
                     traceit("Failed to send un-link command\n");
                     strcpy(nak, "NAK\n");
                  }
                  else
                  {
                     if (ANNOUNCE_LINK)
                     {
                        snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_unlinked.dat",
                                 ANNOUNCE_DIR_NOTIFY, call[0]);
                        post_notification(ANNOUNCE_NOTIFY_FILE);
                     }
                     print_links_file();
                  }
               }
            }
         }
         else
         if ( (strcmp(cmd, "lm") == 0) || (strcmp(cmd, "plm") == 0) ) 
         /* link our module to a reflector module, example: lm BXRF010C or plm BXRF010C */
         {
            if (call[7] == ' ')
            {
               p = strchr(call, ' ');
               if (p)
               {
                  call[7] = *(p - 1);
                  *(p - 1) = ' ';
               }
            }
            if (strlen(call) == 8)
            {
               if (call[0] == 'A')
                  i = 0;
               else
               if (call[0] == 'B')
                  i = 1;
               else
               if (call[0] == 'C')
                  i = 2;
               else
               {
                  i = -1;
                  traceit("The from module %c is invalid\n", call[0]); 
                  strcpy(nak, "NAK\n");
               }

               if (i >= 0)
               {
                  if (mod_links[i].to_mod == '\0')
                  {
                     if ((call[7] == 'A') || (call[7] == 'B') || (call[7] == 'C'))
                     {
                        mod_links[i].from_mod = call[0];
                        memcpy(mod_links[i].to_call, call + 1, 6);
                        mod_links[i].to_call[6] = ' ';
                        mod_links[i].to_call[7] = ' ';
                        mod_links[i].to_call[CALL_SIZE] = '\0';
                        mod_links[i].to_mod = call[7];

                        if (strcmp(cmd, "lm") == 0)
                           result = link_command(i,false);
                        else
                           result = link_command(i,true);

                        if (!result)
                        {
                           if (ANNOUNCE_LINK)
                           {
                              snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_failed_linked.dat",
                                       ANNOUNCE_DIR_NOTIFY, call[0]);
                              post_notification(ANNOUNCE_NOTIFY_FILE);
                           } 
                           traceit("Failed to send link command\n");
                           strcpy(nak, "NAK\n");
                        }
                        else
                        {
                           if (!mod_links[i].persist)
                           {
                              if (ANNOUNCE_LINK)
                              {
                                 strcpy(linked_remote_system, mod_links[i].to_call);
                                 for (j = 0; j < CALL_SIZE; j++)
                                 {
                                    if (linked_remote_system[j] == ' ')
                                       linked_remote_system[j] = '\0';
                                 }
                                 snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_linked.dat_LINKED_%s_%c",
                                          ANNOUNCE_DIR_NOTIFY, call[0],
                                          linked_remote_system, mod_links[i].to_mod);
                                 post_notification(ANNOUNCE_NOTIFY_FILE);
                              }
                           }
                           print_links_file();
                        }
                     } 
                     else
                     {
                        traceit("The module %c to link to, is invalid\n", call[7]);
                        strcpy(nak, "NAK\n");
                     }
                  }
                  else
                  {
                     if (ANNOUNCE_LINK)
                     {
                        strcpy(linked_remote_system, mod_links[i].to_call);
                        for (j = 0; j < CALL_SIZE; j++)
                        {
                           if (linked_remote_system[j] == ' ')
                              linked_remote_system[j] = '\0';
                        }
                        snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_already_linked.dat_LINKED_%s_%c",
                                 ANNOUNCE_DIR_NOTIFY, mod_links[i].from_mod,
                                 linked_remote_system, mod_links[i].to_mod);
                        post_notification(ANNOUNCE_NOTIFY_FILE);
                     }
                     traceit("Module %c is already linked to %s %c,...unlink first\n",
                             mod_links[i].from_mod, mod_links[i].to_call,mod_links[i].to_mod);
                     strcpy(nak, "NAK\n");
                  }
               }
            }
            else
            {
               traceit("Invalid command\n");
               strcpy(nak, "NAK\n");
            }
         } 
         else
         if (strcmp(cmd, "mc") == 0) /* mute a user by callsign */
         {
            if (!mute_call(call, true))
               strcpy(nak, "NAK\n");
         }
         else
         if (strcmp(cmd, "uc") == 0) /* un-mute a user by callsign */
         {
            if (!mute_call(call, false))
               strcpy(nak, "NAK\n");
         }
         else
         {
            traceit("Unknown command\n", cmd);
            strcpy(nak, "NAK\n");
         }
      }
   }

   if (nak[0] == '\0')
      strcpy(nak, "OK\n");
      
   sendto(cmd_sock,nak,4,
          0, (struct sockaddr *)&fromCmd,
          sizeof(struct sockaddr_in));
   return;
}

/* Print administrators */
static void print_admin()
{
   admin_type::iterator pos;
   char buf[32];

   for (pos = admin.begin(); pos != admin.end(); pos++)
   {
      snprintf(buf, 31, "[%s]\n", pos->c_str());
      sendto(cmd_sock, buf, strlen(buf),
             0, (struct sockaddr *)&fromCmd,
             sizeof(struct sockaddr_in));
   }
}

static void print_links_file()
{ 
   struct tm tm1;
   struct tm tm2;
   short i;
   statusfp = fopen(STATUS_FILE, "w");
   if (!statusfp)
      traceit("Failed to create status file %s\n", STATUS_FILE);
   else
   {
      setvbuf(statusfp, (char *)NULL, _IOLBF, 0);
      for (i = 0; i < 3; i++)
      {
         if (mod_links[i].to_mod != '\0')
         {
            localtime_r(&(mod_links[i].link_time),&tm1);
            localtime_r(&(mod_links[i].confirm_time),&tm2);
            fprintf(statusfp, 
                    "%c,%s,%c,%s,%c,%02d%02d%02d,%02d:%02d:%02d,%02d:%02d:%02d\n",
                    mod_links[i].from_mod,
                    mod_links[i].to_call,
                    mod_links[i].to_mod,
                    mod_links[i].to_ip,
                    mod_links[i].persist?'Y':'N',
                    tm1.tm_mon+1,tm1.tm_mday,tm1.tm_year % 100,
                    tm1.tm_hour,tm1.tm_min,tm1.tm_sec,
                    tm2.tm_hour,tm2.tm_min,tm2.tm_sec);
         }
      }
      fclose(statusfp);
   }
}

/* print the links on the screen */
static void print_links_screen()
{
   char buf[128];
   int i;
   struct tm tm1;
   struct tm tm2;

   for (i = 0; i < 3; i++)
   {
      if (mod_links[i].to_mod != '\0')
      {
         localtime_r(&(mod_links[i].link_time),&tm1);
         localtime_r(&(mod_links[i].confirm_time),&tm2);
         snprintf(buf, 127, 
                  "%c,%s,%c,%s,%c,%02d%02d%02d,%02d:%02d:%02d,%02d:%02d:%02d\n",
                  mod_links[i].from_mod,
                  mod_links[i].to_call,
                  mod_links[i].to_mod,
                  mod_links[i].to_ip,
                  mod_links[i].persist?'Y':'N',
                  tm1.tm_mon+1,tm1.tm_mday,tm1.tm_year % 100,
                  tm1.tm_hour,tm1.tm_min,tm1.tm_sec,
                  tm2.tm_hour,tm2.tm_min,tm2.tm_sec);

         sendto(cmd_sock, buf, strlen(buf),
             0, (struct sockaddr *)&fromCmd,
             sizeof(struct sockaddr_in));
      }
   }
}

static void print_version()
{
   char buf[24];
   
   snprintf(buf, 23, "%s\n", VERSION);
   sendto(cmd_sock, buf, strlen(buf),
          0, (struct sockaddr *)&fromCmd,
          sizeof(struct sockaddr_in));
}

/* Print connected users */
static void print_users()
{
   a_user_list_type::iterator pos;
   struct a_user *a_user_ptr;
   struct tm tm;

   char buf[124]; 
   for (pos = a_user_list.begin(); pos != a_user_list.end(); pos++)
   {
      a_user_ptr = (struct a_user *)pos->second;
      localtime_r(&(a_user_ptr->connect_time),&tm);
      snprintf(buf, 123, "%s,%s,%s,%02d%02d%02d,%02d:%02d:%02d,%s\n",
               a_user_ptr->call,
               pos->first.c_str(),
               a_user_ptr->is_dongle?"DONGLE":"REFLECTOR",
               tm.tm_mon+1,tm.tm_mday,tm.tm_year % 100,
               tm.tm_hour,tm.tm_min,tm.tm_sec,
               a_user_ptr->isMute?"Muted":"notMuted");
    
      sendto(cmd_sock, buf, strlen(buf),
             0, (struct sockaddr *)&fromCmd,
             sizeof(struct sockaddr_in));
   }
}
/* Mute or unmute all users */
static void mute_users(bool mute)
{
   a_user_list_type::iterator pos;
   struct a_user *a_user_ptr;

   for (pos = a_user_list.begin(); pos != a_user_list.end(); pos++)
   {
      a_user_ptr = (struct a_user *)pos->second;
      a_user_ptr->isMute = mute;
   }
}

/* mute or unmute a user by callsign */
static bool mute_call(char *call, bool mute)
{
   bool found = false;
   a_user_list_type::iterator pos;
   struct a_user *a_user_ptr;

   for (pos = a_user_list.begin(); pos != a_user_list.end(); pos++)
   {
      a_user_ptr = (struct a_user *)pos->second;
      if (strcmp(a_user_ptr->call, call) == 0)
      {
         found = true;
         a_user_ptr->isMute = mute;
      }
   }
   return found;
}

/* print blocked users */
static void print_blocks()
{
   blocks_type::iterator pos;
   char buf[32];

   for (pos = blocks.begin(); pos != blocks.end(); pos++)
   {
      snprintf(buf, 31, "[%s]\n",  pos->c_str());
      sendto(cmd_sock, buf, strlen(buf),
             0, (struct sockaddr *)&fromCmd,
             sizeof(struct sockaddr_in));
   }
}

/* signal catching function */
static void sigCatch(int signum)
{
   /* do NOT do any serious work here */
   /* Just let dextra_srv get out of its processing loop */

      if ((signum == SIGTERM) ||
       (signum == SIGHUP)  ||
       (signum == SIGQUIT) ||
       (signum == SIGINT))
         keep_running = false;
   return;
}

/* run the main loop for dextra_srv */
static void runit()
{
   int recvlen;
   int recvlen2;
   int recvlen3;
   socklen_t fromlen;

   struct iphdr *ip_hdr;
   struct udphdr *udp_hdr;

   bool allowed_to_connect = true;

   char call[CALL_SIZE + 1];
   char ip[IP_SIZE + 1];
   struct a_user *a_user_ptr;

   /* for the connected station */
   a_user_list_type::iterator user_pos;
   pair<a_user_list_type::iterator,bool> user_insert_pair;

   struct in_addr inaddr;
   char source_mod = ' ';

   char temp[8];
   bool deleted = false;
   char host[MAXHOSTNAMELEN + 1];
   short i,j;

   admin_type::iterator admin_pos;
   blocks_type::iterator block_pos;

   bool result = false;

   /* 
      Data sent from dextra_srv to dextra_reflect to request a link.
      Example: "DB0SAT space space BC NULL" 
               This means that DB0SAT module B wants to link with 
               a reflector module C

          The reply to the link request should be this:
               DB0SAT space space BC ACK NULL
               DB0SAT space space BC NAK NULL

      Data sent from dextra_srv to dextra_reflect to unlink.
      Example: DB0SAT space space B space NULL
               This means that DB0SAT module B wants to unlink
   */

   time(&HBinterStart);
   time(&ConfirmInterStart);

   for (i = 0; i < 3; i++)
      time(&last_active_to_usr_time[i]);

   while (keep_running)
   {
      time(&tNow);

      /* Every 1 second, send heartbeat to connected systems */
      if ((tNow - HBinterStart) > 0)
      {
         send_heartbeat();
         time(&HBinterStart);
      }

      /* every 60 seconds, check for losing the permament connection */
      if ((tNow - ConfirmInterStart) > 60)
      {
         check_confirm();
         time(&ConfirmInterStart);
      }

      for (i = 0; i < 3; i++)
      {
         if ((tNow - last_active_to_usr_time[i]) > 2)
         {
            if (streamid_from_gwy_list[i].size() > 0)
            {
               streamid_from_gwy_list[i].clear();
               traceit("Inactivity...destroyed streamid_from_gwy_list[%d]\n", i);
            }
         }
      }

      /* wait 20 ms max */
      FD_ZERO(&fdset);
      FD_SET(srv_sock,&fdset);
      FD_SET(internal_raw_sock,&fdset);
      FD_SET(external_raw_sock,&fdset);
      FD_SET(cmd_sock,&fdset);
      tv.tv_sec = 0;
      tv.tv_usec = 20000; /* 20 ms */
      (void)select(FD_SETSIZE,&fdset,0,0,&tv);

      /* process shell commands */
      if (FD_ISSET(cmd_sock, &fdset))
      {
         fromlen = sizeof(struct sockaddr_in);
         recvlen3 = recvfrom(cmd_sock,(char *)readBuffer3,124,
                            0,(struct sockaddr *)&fromCmd,&fromlen);
         if (recvlen3 > 0)
         {
            readBuffer3[recvlen3 - 1] = '\0';
            handle_cmd((char *)readBuffer3);
         }

         FD_CLR (cmd_sock,&fdset);
      }

      /* Process g2 */
      if (FD_ISSET(external_raw_sock, &fdset))
      {
         fromlen = sizeof(struct sockaddr_ll);
         recvlen2 = recvfrom(external_raw_sock,(char *)readBuffer2,124,
                             0,(struct sockaddr *)&sll,&fromlen);

         if  (sll.sll_pkttype == PACKET_HOST)
         {
            ip_hdr = (struct iphdr *)(readBuffer2);
            if ((ip_hdr->protocol == IPPROTO_UDP) &&
                !(ntohs(ip_hdr->frag_off) & IP_MF) &&
                (ip_hdr->version == 4))
            {
               udp_hdr = (struct udphdr*)(readBuffer2 + ip_hdr->ihl*4);
               if (udp_hdr->dest == net_GW_PORT)
               {
                  udp_payload_len2 = ntohs(ip_hdr->tot_len) - ip_hdr->ihl*4 - sizeof(struct udphdr);
                  udp_payload2 = (u_char *)(readBuffer2 + ip_hdr->ihl*4 + sizeof(struct udphdr));

                  if ( ((udp_payload_len2 == 56) || 
                        (udp_payload_len2 == 27)) &&
                       (memcmp(udp_payload2, "DSVT", 4) == 0) &&
                       ((udp_payload2[4] == 0x10) ||         /* header */
                        (udp_payload2[4] == 0x20)) &&        /* details */
                       (udp_payload2[8] == 0x20) )           /* voice type */
                  {
                     if (QSO_DETAILS) 
                     {
                        if (udp_payload_len2 == 56)
                        {
                           inaddr.s_addr = ip_hdr->saddr;
                           traceit("START g2: streamID=%d,%d, flags=%02x:%02x:%02x, my=%.8s, sfx=%.4s, ur=%.8s, rpt1=%.8s, rpt2=%.8s, %d bytes fromIP=%s\n",
                                      udp_payload2[12],udp_payload2[13],
                                      udp_payload2[15], udp_payload2[16], udp_payload2[17],
                                      &udp_payload2[42],
                                      &udp_payload2[50], &udp_payload2[34],
                                      &udp_payload2[18], &udp_payload2[26],
                                      udp_payload_len2,inet_ntoa(inaddr));
                        }
                        else
                        {
                           if ((udp_payload2[14] & 0x40) != 0)
                           {
                              inaddr.s_addr = ip_hdr->saddr;
                              traceit("END g2: streamID=%d,%d, %d bytes from IP=%s\n",
                                       udp_payload2[12],udp_payload2[13],udp_payload_len2,inet_ntoa(inaddr));
                           }
                        }
                     }
                     // Must not send, possibility of a loop:  send_a_pkt_to_all(...);                  
                  }
               }
            }
         }
         FD_CLR (external_raw_sock,&fdset);
      }

      /* Process repeater controller packets */
      if (FD_ISSET(internal_raw_sock, &fdset))
      {
         fromlen = sizeof(struct sockaddr_ll);
         recvlen = recvfrom(internal_raw_sock,(char *)readBuffer,124,
                            0,(struct sockaddr *)&sll,&fromlen); 

         /* capture incoming packets from local RF */
         if (sll.sll_pkttype == PACKET_HOST)
         {
               ip_hdr = (struct iphdr *)(readBuffer);
               if ((ip_hdr->protocol == IPPROTO_UDP) &&
                   !(ntohs(ip_hdr->frag_off) & IP_MF) &&
                   (ip_hdr->version == 4) &&
                   (ip_hdr->daddr == INTERNAL_DADDR))
               {

                  /* from repeater controller port ONLY */
                  udp_hdr = (struct udphdr*)(readBuffer + ip_hdr->ihl*4);
                  if ((udp_hdr->source == net_CTRL_PORT) &&
                      (udp_hdr->dest == net_G2_INTERNAL_PORT))
                  {
                     /* Over the UDP header */
                     udp_payload_len = ntohs(ip_hdr->tot_len) - ip_hdr->ihl*4 - sizeof(struct udphdr);

                     /* Point to the UDP payload */
                     udp_payload = (u_char *)(readBuffer + ip_hdr->ihl*4 + sizeof(struct udphdr));

                     if ( ((udp_payload_len == 58) || 
                            (udp_payload_len == 29) || 
                            (udp_payload_len == 32)) &&
                          (memcmp(udp_payload, "DSTR", 4) == 0) &&
                          (udp_payload[6] == 0x73) &&          /* not a reply ? */
                          (udp_payload[10] == 0x20) &&         /* Voice type */
                          (udp_payload[8] == 0x00) &&
                          ((udp_payload[9] == 0x30) ||         /* 48 bytes follow */
                           (udp_payload[9] == 0x13) ||         /* 19 bytes follow */
                           (udp_payload[9] == 0x16)) )         /* 22 bytes follow */
                     {
                        /* Examine CQ call packet */
                        if (udp_payload_len == 58) 
                        {
                           /* 
                              From a local repeater radio user to repeater controller, set:
                              rpt1 <--- rpt2, 
                              rpt2 <--- rpt1 
                           */

                           memcpy(temp, udp_payload + 20, 8);
                           memcpy(udp_payload + 20, udp_payload + 28, 8);
                           memcpy(udp_payload + 28, temp, 8);

                           /* Is this a request to link */
                           if (((udp_payload[43] == LINK_CODE) || (udp_payload[43] == RELINK_CODE)) &&  
                               ((udp_payload[42] == 'A') || (udp_payload[42] == 'B') || (udp_payload[42] == 'C')))
                           {

                              /* possible request to link */
                              /* find out if radio user is allowed to link */
                              memcpy(call, udp_payload + 44, CALL_SIZE);
                              call[CALL_SIZE] = '\0';

                              admin_pos = admin.find(call);
                              if ((admin_pos != admin.end()) || LINK_BY_USERS)
                              {
                                 traceit("link request %.8s was captured at local repeater\n",udp_payload + 36);
                                 /* rpt1 must contain module A or module B or module C */
                                 if (udp_payload[27] == 'A')
                                    i = 0;
                                 else
                                 if (udp_payload[27] == 'B')
                                    i = 1;
                                 else
                                 if (udp_payload[27] == 'C')
                                    i = 2;
                                 else
                                 {
                                    i = -1;
                                    traceit("RPT1 module info %c invalid for linking\n", udp_payload[27]);
                                 }

                                 if (i >= 0)
                                 {
                                    /* Is this particular module linked already ? */
                                    if (mod_links[i].to_mod == '\0')
                                    {
                                       /* Try to link one of our repeater modules to a reflector module */
                                       memcpy(mod_links[i].to_call, udp_payload + 36, CALL_SIZE - 2);
                                       mod_links[i].to_call[6] = ' ';
                                       mod_links[i].to_call[7] = ' ';
                                       mod_links[i].to_call[CALL_SIZE] = '\0';

                                       mod_links[i].from_mod = udp_payload[27];
                                       mod_links[i].to_mod = udp_payload[42];

                                       if (udp_payload[43] == LINK_CODE)
                                          result = link_command(i,false);
                                       else
                                          result = link_command(i, true);

                                       if (!result)
                                       {
                                          if (ANNOUNCE_LINK)
                                          {
                                             snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_failed_linked.dat",
                                                      ANNOUNCE_DIR_NOTIFY, udp_payload[27]);
                                             post_notification(ANNOUNCE_NOTIFY_FILE);
                                          }
                                          traceit("Failed to send link command\n");
                                       }
                                       else
                                       {
                                          if (!mod_links[i].persist)
                                          {
                                             if (ANNOUNCE_LINK)
                                             {
                                                strcpy(linked_remote_system, mod_links[i].to_call);
                                                for (j = 0; j < CALL_SIZE; j++)
                                                {
                                                   if (linked_remote_system[j] == ' ')
                                                      linked_remote_system[j] = '\0';
                                                }
                                                snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_linked.dat_LINKED_%s_%c",
                                                         ANNOUNCE_DIR_NOTIFY, udp_payload[27],
                                                         linked_remote_system, mod_links[i].to_mod);
                                                post_notification(ANNOUNCE_NOTIFY_FILE);
                                             }
                                          }
                                          print_links_file();
                                       }
                                    }
                                    else
                                    {
                                       if (ANNOUNCE_LINK)
                                       {
                                          strcpy(linked_remote_system, mod_links[i].to_call);
                                          for (j = 0; j < CALL_SIZE; j++)
                                          {
                                             if (linked_remote_system[j] == ' ')
                                                linked_remote_system[j] = '\0';
                                          }
                                          snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX,"%s/%c_already_linked.dat_LINKED_%s_%c",
                                                   ANNOUNCE_DIR_NOTIFY, mod_links[i].from_mod,
                                                   linked_remote_system, mod_links[i].to_mod);
                                          post_notification(ANNOUNCE_NOTIFY_FILE);
                                       }
                                       traceit("Module %c is already linked to %s %c,...unlink first\n",
                                               udp_payload[27], mod_links[i].to_call, mod_links[i].to_mod); 
                                    }
                                 }
                              }
                           }
                           else
                           if (udp_payload[43] == IS_IT_LINKED_CODE)
                           {
                              /* radio user wants to know if a particular module is linked */
                              memcpy(call, udp_payload + 44, CALL_SIZE);
                              call[CALL_SIZE] = '\0';

                              admin_pos = admin.find(call);
                              if ((admin_pos != admin.end()) || LINK_BY_USERS)
                              {
                                 traceit("LinkINFO CODE %.8s was captured at local repeater\n",udp_payload + 36); 
                                 /* rpt1 must contain module A or module B or module C */
                                 if (udp_payload[27] == 'A')
                                    i = 0;
                                 else
                                 if (udp_payload[27] == 'B')
                                    i = 1;
                                 else
                                 if (udp_payload[27] == 'C')
                                    i = 2;
                                 else
                                 {
                                    i = -1;
                                    traceit("RPT1 module info %c invalid for linkINFO\n", udp_payload[27]);
                                 } 
                                 if (i >= 0)
                                 {
                                    if (ANNOUNCE_LINK)
                                    {
                                       if (mod_links[i].to_mod != '\0')
                                       { 
                                          strcpy(linked_remote_system, mod_links[i].to_call);
                                          for (j = 0; j < CALL_SIZE; j++)
                                          {
                                             if (linked_remote_system[j] == ' ')
                                                linked_remote_system[j] = '\0';
                                          }
                                          snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_linked.dat_LINKED_%s_%c",
                                                   ANNOUNCE_DIR_NOTIFY, udp_payload[27],
                                                   linked_remote_system, mod_links[i].to_mod);
                                          post_notification(ANNOUNCE_NOTIFY_FILE);
                                       }
                                       else
                                       {
                                          snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_unlinked.dat",
                                                   ANNOUNCE_DIR_NOTIFY, udp_payload[27]);
                                          post_notification(ANNOUNCE_NOTIFY_FILE);
                                       }
                                    }
                                 }
                              }
                           }
                           else
                           if (udp_payload[43] == UNLINK_CODE) 
                           {
                              /* possible request to unlink */
                              /* find out if radio user is allowed to unlink */
                              memcpy(call, udp_payload + 44, CALL_SIZE);
                              call[CALL_SIZE] = '\0';

                              admin_pos = admin.find(call);
                              if ((admin_pos != admin.end()) || LINK_BY_USERS)
                              {
                                 traceit("Un-link request %.8s was captured at local repeater\n",udp_payload + 36);
                                 /* rpt1 must contain module A or module B or module C */
                                 if (udp_payload[27] == 'A')
                                    i = 0;
                                 else
                                 if (udp_payload[27] == 'B')
                                    i = 1;
                                 else
                                 if (udp_payload[27] == 'C')
                                    i = 2;
                                 else
                                 {
                                    i = -1;
                                    traceit("RPT1 module info %c invalid for un-linking\n", udp_payload[27]);
                                 }
                                 if (i >= 0)
                                 {
                                    if (mod_links[i].to_mod == '\0') 
                                    {
                                       traceit("Module %c is already un-linked\n", udp_payload[27]);

                                       if (ANNOUNCE_LINK)
                                       {
                                          snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX,"%s/%c_already_unlinked.dat",
                                                   ANNOUNCE_DIR_NOTIFY,udp_payload[27]);
                                          post_notification(ANNOUNCE_NOTIFY_FILE);
                                       }
                                    }
                                    else
                                    {
                                       if (!unlink_command(i))
                                          traceit("Failed to send unlink command\n");
                                       else
                                       {
                                          if (ANNOUNCE_LINK)
                                          {
                                             snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_unlinked.dat",
                                                      ANNOUNCE_DIR_NOTIFY, udp_payload[27]);
                                             post_notification(ANNOUNCE_NOTIFY_FILE);
                                          }
                                          print_links_file();
                                       }
                                    }
                                 }
                              }
                           }
                        }
                      
                        if (udp_payload_len == 58)
                        {
                           if (QSO_DETAILS)
                              traceit("START capture: streamID=%d,%d, flags=%02x:%02x:%02x, my=%.8s, sfx=%.4s, ur=%.8s, rpt1=%.8s, rpt2=%.8s, %d bytes\n",
                                       udp_payload[14], udp_payload[15],
                                       udp_payload[17], udp_payload[18], udp_payload[19],
                                       udp_payload + 44, udp_payload + 52, udp_payload + 36,
                                       udp_payload + 20, udp_payload + 28,
                                       udp_payload_len);
                        }
                        else
                        {
                           if ((udp_payload[16] & 0x40) != 0)
                           {
                              if (QSO_DETAILS)
                                 traceit("END capture: streamID=%d,%d, %d bytes\n",
                                         udp_payload[14], udp_payload[15], udp_payload_len);
                           }
                        }
                        prepare_rawbytes(udp_payload, udp_payload_len);   
                        send_a_pkt_to_all(rawbytes,rawbytes_len, 0);
                     }
                  }                     
               }
         }
         FD_CLR (internal_raw_sock,&fdset);
      }

      if (FD_ISSET(srv_sock, &fdset))
      {
         fromlen = sizeof(struct sockaddr_in);
         recvlen = recvfrom(srv_sock,(char *)readBuffer,124,
                            0,(struct sockaddr *)&fromUser,&fromlen);

         strncpy(ip, inet_ntoa(fromUser.sin_addr),IP_SIZE);
         ip[IP_SIZE] = '\0';

         /* possible ACK or NAK from remote system, 
            either link request was accepted by the remote system
            or it was rejected.

          The reply to the link request should be this:
               DB0SAT space space BC ACK NULL
               DB0SAT space space BC NAK NULL

         */
         if (recvlen == (CALL_SIZE + 6))
         {
            if (memcmp(OWNER, (char *)readBuffer, CALL_SIZE) == 0)
            {
               for (i = 0; i < 3; i++)
               {
                  if ((strcmp(ip, mod_links[i].to_ip) == 0) && 
                      (readBuffer[8] == mod_links[i].from_mod) &&
                      (readBuffer[9] == mod_links[i].to_mod))
                  {
                     if (memcmp((char *)readBuffer + 10, "ACK", 3) == 0)
                     {
                        traceit("Link module %c to %s %c is now confirmed\n", 
                                mod_links[i].from_mod,
                                mod_links[i].to_call,
                                mod_links[i].to_mod);

                        time(&(mod_links[i].confirm_time));

                        /* 
                           The link request was accepted.
                           Now add the remote station to the list, if it is NOT there.
                           We do not need to wait for the KEEPALIVE from the remote system,
                             because the KEEPALIVE may get lost.
                        */

                        user_pos = a_user_list.find(ip);
                        if (user_pos != a_user_list.end())
                        {
                           /* remote station is already on our list */
                           if (mod_links[i].persist)
                           {
                              if (ANNOUNCE_LINK)
                              {
                                 strcpy(linked_remote_system, mod_links[i].to_call);
                                 for (j = 0; j < CALL_SIZE; j++)
                                 {
                                    if (linked_remote_system[j] == ' ')
                                       linked_remote_system[j] = '\0';
                                 }
                                 snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_linked.dat_LINKED_%s_%c",
                                          ANNOUNCE_DIR_NOTIFY, mod_links[i].from_mod,
                                          linked_remote_system, mod_links[i].to_mod);
                                 post_notification(ANNOUNCE_NOTIFY_FILE);
                              }
                           }
                        }
                        else
                        {
                           /* remote station is NOT on our list, add it */
                           a_user_ptr = (struct a_user *)malloc(sizeof(struct a_user));
                           if (a_user_ptr)
                           {
                              a_user_ptr->countdown = TIMEOUT;
                              memcpy((char *)&(a_user_ptr->sin), (char *)&fromUser, sizeof(struct sockaddr_in));
                              strcpy(a_user_ptr->call, mod_links[i].to_call);
                              a_user_ptr->isMute = false;
                              a_user_ptr->sin.sin_port = htons(LISTEN_PORT);
                              a_user_ptr->is_dongle = false;
                              a_user_ptr->connect_time = mod_links[i].confirm_time;
                              user_insert_pair = a_user_list.insert(pair<string, struct a_user *>(ip, a_user_ptr));
                              if (user_insert_pair.second)
                              {
                                 traceit("new CALL=%s,  %s,  ip=%s, users=%d\n",
                                         a_user_ptr->call,
                                         (a_user_ptr->is_dongle)?"DONGLE":"REFLECTOR",
                                         ip, a_user_list.size());

                                if (mod_links[i].persist)
                                {
                                   if (ANNOUNCE_LINK)
                                   {
                                      strcpy(linked_remote_system, mod_links[i].to_call);
                                      for (j = 0; j < CALL_SIZE; j++)
                                      {
                                         if (linked_remote_system[j] == ' ')
                                            linked_remote_system[j] = '\0';
                                      }
                                      snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_linked.dat_LINKED_%s_%c",
                                               ANNOUNCE_DIR_NOTIFY, mod_links[i].from_mod,
                                               linked_remote_system, mod_links[i].to_mod);
                                      post_notification(ANNOUNCE_NOTIFY_FILE);
                                   }
                                }
                              }
                              else
                              {
                                 traceit("failed to add CALL=%s,  ip=%s\n",a_user_ptr->call,ip);
                                 free(a_user_ptr);
                                 a_user_ptr = NULL;

                                 mod_links[i].to_ip[0] = '\0';
                                 mod_links[i].to_call[0] = '\0';
                                 mod_links[i].from_mod = '\0';
                                 mod_links[i].to_mod = '\0';
                                 mod_links[i].link_time = 0;
                                 mod_links[i].confirm_time = 0;
                                 mod_links[i].persist = false;
                              }
                           }
                           else
                           {
                              traceit("malloc() failed for call=%s,  ip=%s\n",mod_links[i].to_call,ip);

                              mod_links[i].to_ip[0] = '\0';
                              mod_links[i].to_call[0] = '\0';
                              mod_links[i].from_mod = '\0';
                              mod_links[i].to_mod = '\0';
                              mod_links[i].link_time = 0;
                              mod_links[i].confirm_time = 0;
                              mod_links[i].persist = false;
                           }
                        }
                     }
                     else
                     if (memcmp((char *)readBuffer + 10, "NAK", 3) == 0)
                     {
                        traceit("Link module %c to %s %c is rejected,...dropping link\n",
                                mod_links[i].from_mod,
                                mod_links[i].to_call,
                                mod_links[i].to_mod);

                        mod_links[i].from_mod = '\0';
                        mod_links[i].to_mod = '\0';
                        mod_links[i].to_ip[0] = '\0';
                        mod_links[i].to_call[0] = '\0';
                        mod_links[i].link_time = 0;
                        mod_links[i].confirm_time = 0;
                        mod_links[i].persist = false;

                        if (ANNOUNCE_LINK)
                        {
                           snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_unlinked.dat",
                                    ANNOUNCE_DIR_NOTIFY, readBuffer[8]);
                           post_notification(ANNOUNCE_NOTIFY_FILE);
                        }
                     }
                     print_links_file();
                     break;
                  }
               }
            }
         }
         else
         if ((recvlen == (CALL_SIZE + 1)) || (recvlen == (CALL_SIZE + 3)))
         {
            strncpy(call, (char *)readBuffer,CALL_SIZE);
            call[CALL_SIZE] = '\0';

            /* is this a new user? */
            user_pos = a_user_list.find(ip);
            if (user_pos != a_user_list.end())
            {
               deleted = false;

               a_user_ptr = (struct a_user *)user_pos->second;
               a_user_ptr->countdown = TIMEOUT;

               memcpy((char *)&(a_user_ptr->sin),(char *)&fromUser, sizeof(struct sockaddr_in));

               /* If not a dongle */
               if (!a_user_ptr->is_dongle) 
                  a_user_ptr->sin.sin_port = htons(LISTEN_PORT);

               /* Same IP address, but call changed ? */
               if (memcmp(a_user_ptr->call, call, CALL_SIZE) != 0)
               {
                  traceit("Same IP %s but call changed from %s to %s, will remove old call\n",
                          ip, a_user_ptr->call, call);

                  strcpy(call,a_user_ptr->call);

                  for (i = 0; i < 3; i++)
                  {
                     if (strcmp(mod_links[i].to_ip, ip) == 0)
                     {
                        if (ANNOUNCE_LINK)
                        {
                           snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX,"%s/%c_unlinked.dat",
                                    ANNOUNCE_DIR_NOTIFY, mod_links[i].from_mod);
                           post_notification(ANNOUNCE_NOTIFY_FILE);
                        }

                        mod_links[i].from_mod = '\0';
                        mod_links[i].to_mod = '\0';
                        mod_links[i].to_call[0] = '\0';
                        mod_links[i].to_ip[0] = '\0';
                        mod_links[i].link_time = 0;
                        mod_links[i].confirm_time = 0;
                        mod_links[i].persist = false;

                        print_links_file();
                     }
                  }
                  free(user_pos->second);
                  a_user_ptr = user_pos->second = NULL;
                  a_user_list.erase(user_pos);
                  deleted = true;

                  traceit("Deleted old record %s\n", call);
               }

               /* User has been blocked ?, if yes, remove */
               if (!deleted)
               {
                  block_pos = blocks.find(call);
                  if (block_pos != blocks.end())
                  {
                     traceit("Call %s with %s will be removed because it is now blocked\n",call,ip);
                     for (i = 0; i < 3; i++)
                     {
                        if (strcmp(mod_links[i].to_ip, ip) == 0)
                        {
                           if (ANNOUNCE_LINK)
                           {
                              snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX,"%s/%c_unlinked.dat",
                                       ANNOUNCE_DIR_NOTIFY, mod_links[i].from_mod);
                              post_notification(ANNOUNCE_NOTIFY_FILE);
                           }

                           mod_links[i].from_mod = '\0';
                           mod_links[i].to_mod = '\0';
                           mod_links[i].to_call[0] = '\0';
                           mod_links[i].to_ip[0] = '\0';
                           mod_links[i].link_time = 0;
                           mod_links[i].confirm_time = 0;
                           mod_links[i].persist = false;

                           print_links_file();
                        }
                     }
                     free(user_pos->second);
                     a_user_ptr = user_pos->second = NULL;
                     a_user_list.erase(user_pos);
                     deleted = true;
                  }
               }

               if (!deleted)
               {
                  /* Update from another repeater ? */
                  if ((recvlen == (CALL_SIZE + 3)) && (!a_user_ptr->is_dongle))
                  {
                     traceit("Possible new link/unlink request received: %.*s\n", recvlen - 1, readBuffer);
                     if ((readBuffer[8] == 'A') || (readBuffer[8] == 'B') || (readBuffer[8] == 'C'))
                     {
                        /* Un-link request ? */
                        if (readBuffer[9] == ' ')
                        {
                           /* find out if the module is linked */
                           for (i = 0; i < 3; i++)
                           {
                              if ((strcmp(ip, mod_links[i].to_ip) == 0) &&
                                  (readBuffer[8] == mod_links[i].to_mod))
                              {
                                 traceit("Call %s mod %c drops the link\n",call, readBuffer[8]);

                                 if (ANNOUNCE_LINK)
                                 {
                                    snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX,"%s/%c_unlinked.dat",
                                             ANNOUNCE_DIR_NOTIFY, mod_links[i].from_mod);
                                    post_notification(ANNOUNCE_NOTIFY_FILE);
                                 }

                                 mod_links[i].to_mod = '\0';
                                 mod_links[i].to_call[0] = '\0';
                                 mod_links[i].to_ip[0] = '\0';
                                 mod_links[i].from_mod = '\0';
                                 mod_links[i].link_time = 0;
                                 mod_links[i].confirm_time = 0;
                                 mod_links[i].persist = false;

                                 print_links_file();
                                 break;
                              }
                           }

                           /* there was a match and the link was broken */
                           /* if there are no more links to that station, remove it */
                           if (i < 3)
                           {
                              for (i = 0; i < 3; i++)
                              {
                                 if (strcmp(ip, mod_links[i].to_ip) == 0)
                                    break;
                              }
                              if (i > 2)
                              {
                                 free(user_pos->second);
                                 a_user_ptr = user_pos->second = NULL;
                                 a_user_list.erase(user_pos);
                                 deleted = true;
                                 traceit("all links for %s are broken, removing station\n",call);   
                              }
                           }
                        }
                        else
                        {
                           /* Link request from another system */
                           i = -1;
                           if (readBuffer[9] == 'A')
                              i = 0;
                           else
                           if (readBuffer[9] == 'B')
                              i = 1;
                           else
                           if (readBuffer[9] == 'C')
                              i = 2;
                           
                           if (i >= 0)
                           {
                              if (mod_links[i].to_mod == '\0')
                              {
                                 traceit("Link established from %s mod %c ---> mod %c\n",
                                         a_user_ptr->call, readBuffer[8], readBuffer[9]);

                                 mod_links[i].to_mod = readBuffer[8];
                                 strcpy(mod_links[i].to_call, call);  
                                 mod_links[i].from_mod = readBuffer[9];
                                 strcpy(mod_links[i].to_ip, ip);
                                 time(&(mod_links[i].link_time));
                                 time(&(mod_links[i].confirm_time));
                                 mod_links[i].persist = false;

                                 /***
                                    Do not allow another of our repeater modules to
                                    be linked to the same remote system/module
                                 ***/
                                 for (j = 0; j < 3; j++)
                                 {
                                    if (j != i)
                                    {
                                       if ((strcmp(mod_links[j].to_call, mod_links[i].to_call) == 0) &&
                                           (mod_links[j].to_mod == mod_links[i].to_mod))
                                       {
                                          traceit("Old link from %s mod %c to our mod %c, cleared\n",
                                                   mod_links[j].to_call, mod_links[j].to_mod,
                                                   mod_links[j].from_mod);

                                          mod_links[j].to_ip[0] = '\0';
                                          mod_links[j].to_call[0] = '\0';
                                          mod_links[j].from_mod = '\0';
                                          mod_links[j].to_mod = '\0';
                                          mod_links[j].link_time = 0;
                                          mod_links[j].confirm_time = 0;
                                          mod_links[j].persist = false;
                                       }
                                    }
                                 }

                                 if (ANNOUNCE_LINK)
                                 {
                                    strcpy(linked_remote_system, mod_links[i].to_call);
                                    for (j = 0; j < CALL_SIZE; j++)
                                    {
                                       if (linked_remote_system[j] == ' ')
                                          linked_remote_system[j] = '\0';
                                    }
                                    snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_linked.dat_LINKED_%s_%c",
                                             ANNOUNCE_DIR_NOTIFY,mod_links[i].from_mod,
                                             linked_remote_system, mod_links[i].to_mod);
                                    post_notification(ANNOUNCE_NOTIFY_FILE);
                                 }
                                 print_links_file();

                                 memcpy(readBuffer + 10, "ACK", 4);
                                 sendto(srv_sock, readBuffer, CALL_SIZE + 6,
                                        0,(struct sockaddr *)&(a_user_ptr->sin),
                                        sizeof(struct sockaddr_in));
                              }
                              else
                              {
                                 traceit("mod %c is already linked to %s %c\n", 
                                         mod_links[i].from_mod, mod_links[i].to_call, mod_links[i].to_mod);
                              
                                 /* send a NAK here only if our repeater module is linked to something else */
                                 if ((mod_links[i].to_mod != readBuffer[8]) ||
                                     (strcmp(mod_links[i].to_ip, ip) != 0)) 
                                 {
                                    memcpy(readBuffer + 10, "NAK", 4);
                                    sendto(srv_sock, readBuffer, CALL_SIZE + 6,
                                           0,(struct sockaddr *)&(a_user_ptr->sin),
                                           sizeof(struct sockaddr_in));
                                 }
                                 else
                                 {
                                    memcpy(readBuffer + 10, "ACK", 4);
                                    sendto(srv_sock, readBuffer, CALL_SIZE + 6,
                                           0,(struct sockaddr *)&(a_user_ptr->sin),
                                           sizeof(struct sockaddr_in)); 
                                 }
                              }
                           }
                        }
                     }
                  }
               }
            }
            else
            {
               allowed_to_connect = true;
               host[0] = '\0';
               traceit("Possible new connection: %.*s ip=%s\n", recvlen - 1, readBuffer, ip);

               if (a_user_list.size() + 1 > MAX_USERS)
                  allowed_to_connect = false;
               else
               if (memcmp(call,OWNER,CALL_SIZE) == 0)
                  allowed_to_connect = false;
               else
               {
                  block_pos = blocks.find(call);
                  if (block_pos != blocks.end())
                     allowed_to_connect = false;
               }

               if (allowed_to_connect)
               {
                  if (!get_ip(call, host))
                     allowed_to_connect = false;
                  else
                  {
                     if (recvlen == (CALL_SIZE + 1))
                     {
                        if (host[0] != '\0')
                        {
                           /* find out if we requested a link */
                           for (i = 0; i < 3; i++)
                           {
                              if (strcmp(mod_links[i].to_ip,ip) == 0)
                                 break;
                           }
                           if (i > 2)
                           {
                              traceit("%s %s is trying to contact us, but no previous request sent to link to it\n",
                                    call, ip);
                              allowed_to_connect = false;
                           }
                        }
                     }
                     else
                     {
                        if (host[0] == '\0')
                        {
                           traceit("Received link information from %s %s but station does not have a host in db\n",
                                call, ip);
                           allowed_to_connect = false;
                        }
                        else
                        {
                           if ((readBuffer[9] != 'A') && (readBuffer[9] != 'B') && (readBuffer[9] != 'C'))
                              allowed_to_connect = false;
                           else
                           if ((readBuffer[8] != 'A') && (readBuffer[8] != 'B') && (readBuffer[8] != 'C'))
                              allowed_to_connect = false;
                           else
                           {
                              i = -1;
                              if (readBuffer[9] == 'A')
                                 i = 0;
                              else
                              if (readBuffer[9] == 'B')
                                 i = 1;
                              else
                              if (readBuffer[9] == 'C')
                                 i = 2;

                              if (i >= 0)
                              {
                                 if (mod_links[i].to_mod != '\0')
                                 {
                                    traceit("module %c already linked to %s %c\n",
                                             mod_links[i].from_mod,
                                             mod_links[i].to_call, mod_links[i].to_mod);
                                    allowed_to_connect = false;
                                 }
                              }
                           }
                        }
                     }
                  }
               }

               if (allowed_to_connect)
               {
                  /* about to add the new user to the map */
                  a_user_ptr = (struct a_user *)malloc(sizeof(struct a_user));
                  if (a_user_ptr)
                  {
                     a_user_ptr->countdown = TIMEOUT;
                     memcpy((char *)&(a_user_ptr->sin), (char *)&fromUser, sizeof(struct sockaddr_in));
                     strncpy(a_user_ptr->call, call, CALL_SIZE);
                     a_user_ptr->call[CALL_SIZE] = '\0';

                     a_user_ptr->isMute = false;

                     if (host[0] == '\0')
                        a_user_ptr->is_dongle = true;
                     else
                        a_user_ptr->is_dongle = false;

                     /* If not a dongle */
                     if (!a_user_ptr->is_dongle)
                        a_user_ptr->sin.sin_port = htons(LISTEN_PORT);

                     time(&(a_user_ptr->connect_time));

                     user_insert_pair = a_user_list.insert(pair<string, struct a_user *>(ip, a_user_ptr));
                     if (user_insert_pair.second)
                     {
                        traceit("new CALL=%s,%s,ip=%s, users=%d\n",
                                a_user_ptr->call,
                                (a_user_ptr->is_dongle)?"DONGLE":"REFLECTOR", 
                                ip, a_user_list.size());

                        if (recvlen == (CALL_SIZE + 3))
                        {
                           i = -1;
                           if (readBuffer[9] == 'A')
                              i = 0;
                           else
                           if (readBuffer[9] == 'B')
                              i = 1;
                           else
                           if (readBuffer[9] == 'C')
                              i = 2;

                           if (i >= 0)
                           { 
                              traceit("Link established from %s mod %c ---> mod %c\n",
                                      call, readBuffer[8], readBuffer[9]);

                              mod_links[i].to_mod = readBuffer[8];
                              strcpy( mod_links[i].to_call, call);
                              mod_links[i].from_mod = readBuffer[9];
                              strcpy(mod_links[i].to_ip, ip);
                              time(&(mod_links[i].link_time));
                              time(&(mod_links[i].confirm_time));
                              mod_links[i].persist = false;

                              if (ANNOUNCE_LINK)
                              {
                                 strcpy(linked_remote_system, mod_links[i].to_call);
                                 for (j = 0; j < CALL_SIZE; j++)
                                 {
                                    if (linked_remote_system[j] == ' ')
                                       linked_remote_system[j] = '\0';
                                 }
                                 snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_linked.dat_LINKED_%s_%c",
                                          ANNOUNCE_DIR_NOTIFY, mod_links[i].from_mod,
                                          linked_remote_system, mod_links[i].to_mod);
                                 post_notification(ANNOUNCE_NOTIFY_FILE);
                              }
                              print_links_file();

                              memcpy(readBuffer + 10, "ACK", 4);
                              sendto(srv_sock, readBuffer, CALL_SIZE + 6,
                                     0,(struct sockaddr *)&(a_user_ptr->sin),
                                     sizeof(struct sockaddr_in));
                           }
                        }
                        else
                        {
                           /* send HEARTBEAT */
                           sendto(srv_sock, OWNER, strlen(OWNER) + 1,
                                  0,(struct sockaddr *)&(a_user_ptr->sin),
                                  sizeof(struct sockaddr_in));
                        }
                     }
                     else
                     {
                        traceit("failed to add CALL=%s,ip=%s\n",a_user_ptr->call,ip);

                        if (recvlen == (CALL_SIZE + 3))
                        {
                           memcpy(readBuffer + 10, "NAK", 4);
                           sendto(srv_sock, readBuffer, CALL_SIZE + 6,
                                  0,(struct sockaddr *)&(a_user_ptr->sin),
                                  sizeof(struct sockaddr_in));
                        }
                        free(a_user_ptr);
                        a_user_ptr = NULL;

                        for (i = 0; i < 3; i++)
                        {
                           if (strcmp(ip, mod_links[i].to_ip) == 0)
                           {
                              mod_links[i].to_ip[0] = '\0';
                              mod_links[i].to_call[0] = '\0';
                              mod_links[i].from_mod = '\0';
                              mod_links[i].to_mod = '\0';
                              mod_links[i].link_time = 0;
                              mod_links[i].confirm_time = 0;
                              mod_links[i].persist = false;
                              break;
                           }
                        }
                     }
                  }
                  else
                  {
                     traceit("malloc() failed for call=%s,ip=%s\n",call,ip);

                     if (recvlen == (CALL_SIZE + 3))
                     {
                        memcpy(readBuffer + 10, "NAK", 4);

                        fromUser.sin_port = htons(LISTEN_PORT);

                        sendto(srv_sock, readBuffer, CALL_SIZE + 6,
                               0,(struct sockaddr *)&fromUser,
                               sizeof(struct sockaddr_in));
                     }

                     for (i = 0; i < 3; i++)
                     {
                        if (strcmp(ip, mod_links[i].to_ip) == 0)
                        {
                           mod_links[i].to_ip[0] = '\0';
                           mod_links[i].to_call[0] = '\0';
                           mod_links[i].from_mod = '\0';
                           mod_links[i].to_mod = '\0';
                           mod_links[i].link_time = 0;
                           mod_links[i].confirm_time = 0;
                           mod_links[i].persist = false;
                           break;
                        }
                     }
                  }
               }
               else
               {
                  /*
                     If it was a link request, send back a NAK.
                     If it was an unlink request, do nothing
                  */
                  if (recvlen == (CALL_SIZE + 3))
                  {
                     if ((readBuffer[8] != ' ') && (readBuffer[9] == ' '))
                        ;
                     else
                     {
                        memcpy(readBuffer + 10, "NAK", 4);

                        fromUser.sin_port = htons(LISTEN_PORT);

                        sendto(srv_sock, readBuffer, CALL_SIZE + 6,
                               0,(struct sockaddr *)&fromUser,
                               sizeof(struct sockaddr_in));
                     }
                  }

                  /* 
                     For whatever reason, we did NOT add the station to 
                     our list of connected stations, so remove
                     any outstanding link requests 
                  */

                  for (i = 0; i < 3; i++)
                  {
                     if (strcmp(ip, mod_links[i].to_ip) == 0)
                     {
                        mod_links[i].to_ip[0] = '\0';
                        mod_links[i].to_call[0] = '\0';
                        mod_links[i].from_mod = '\0';
                        mod_links[i].to_mod = '\0';
                        mod_links[i].link_time = 0;
                        mod_links[i].confirm_time = 0;
                        mod_links[i].persist = false;
                        break;
                     }
                  }
               }
            }
         }
         else
         if ((recvlen == 56) || (recvlen == 27))
         {
            user_pos = a_user_list.find(ip);
            if (user_pos != a_user_list.end())
            {
               /* point to the connected user */
               a_user_ptr = (struct a_user *)user_pos->second;

               /* refresh timeout */
               a_user_ptr->countdown = TIMEOUT;

               /* If the connected user is MUTED, do not process the audio packets */
               if (!a_user_ptr->isMute)
               {
                  if (recvlen == 56)
                  {
                     if ((!a_user_ptr->is_dongle) &&   /* dongle user ?, do NOTHING */
                         (memcmp(readBuffer + 18, OWNER, 7) != 0)) /* remote repeater converted at the source ?,  do NOTHING  */ 
                     {
                        /* It is a reflector, correct rpt1, rpt2 and re-compute PFCS */
                        for (i = 0; i < 3; i++)
                        {
                           if ((fromUser.sin_addr.s_addr == inet_addr(mod_links[i].to_ip)) &&
                               (readBuffer[25] == mod_links[i].to_mod))
                           {
                              memcpy(&readBuffer[18], OWNER, OWNER_SIZE);
                              readBuffer[25] = mod_links[i].from_mod;
                              memcpy(&readBuffer[26], OWNER, OWNER_SIZE);
                              readBuffer[33] = 'G';
                              memcpy(&readBuffer[34], "CQCQCQ  ", 8);
                              calcPFCS(readBuffer);
                              break;
                           }
                        }
                     }

                     /* At this point, conversions have been completed */
                     if (memcmp(readBuffer + 18, OWNER, 7) == 0)
                     {
                        if (a_user_ptr->is_dongle)
                           source_mod = ' ';
                        else
                        {
                           if (readBuffer[25] == 'A')
                              source_mod = mod_links[0].to_mod;
                           else
                           if (readBuffer[25] == 'B')
                              source_mod = mod_links[1].to_mod;
                           else
                           if (readBuffer[25] == 'C')
                              source_mod = mod_links[2].to_mod;
                           else
                              source_mod = '\0';
                        }
                     }
                     else
                        source_mod = '\0';


                     if (QSO_DETAILS)
                     {
                        if (memcmp(old_streamid, readBuffer + 12, 2) != 0)
                        {
                           memcpy(old_streamid, readBuffer + 12, 2);
                           if (source_mod != '\0')
                           { 
                              traceit("START user: streamID=%d,%d, flags=%02x:%02x:%02x, my=%.8s, sfx=%.4s, ur=%.8s, rpt1=%.8s, rpt2=%.8s, %d bytes fromIP=%s, src=%.8s %c\n",
                                      readBuffer[12], readBuffer[13],
                                      readBuffer[15], readBuffer[16], readBuffer[17],
                                      &readBuffer[42],
                                      &readBuffer[50], &readBuffer[34],
                                      &readBuffer[18], &readBuffer[26],
                                      recvlen,ip,
                                      a_user_ptr->call, source_mod);

                           }
                        }
                     }

                     if (source_mod != '\0')
                     {
                        /* send to local repeater */
                        sendto(srv_sock, (char *)readBuffer, recvlen,
                               0,(struct sockaddr *)&toDst4,sizeof(struct sockaddr_in));

                        /* send to linked stations and connected users */
                        send_a_pkt_to_all(readBuffer,recvlen,fromUser.sin_addr.s_addr); 
                     }
                  }
                  else
                  {
                     if ((readBuffer[14] & 0x40) != 0)
                     {
                        if (QSO_DETAILS)
                        {
                           traceit("END user: streamID=%d,%d, %d bytes from IP=%s\n",
                                   readBuffer[12],readBuffer[13],recvlen,ip);

                           if (memcmp(old_streamid, readBuffer + 12, 2) == 0)
                              memset(old_streamid, 0, 2);
                        }
                     }
                     sendto(srv_sock, (char *)readBuffer, recvlen,
                            0,(struct sockaddr *)&toDst4,sizeof(struct sockaddr_in));

                     /* send to other users */
                     send_a_pkt_to_all(readBuffer,recvlen,fromUser.sin_addr.s_addr);
                  }
               }
            }
         }
         FD_CLR (srv_sock,&fdset);
      }
   }
}

int main(int argc, char **argv)
{
   short i;
   int rc = 0;
   struct sigaction act;

   if (argc != 3)
   {
      printf("Usage:  dextra_srv <configFile> <logFile>\n");
      printf("Example: dextra_srv  ./dextra_srv.cfg  ./dextra_srv.log\n");
      return 1;
   }

   if (access(argv[2], F_OK) == 0)
   {
      printf("Log file %s exists, is dextra_srv already running\n",argv[2]);
      return 1;
   } 

   /* Open log file */
   logfp = fopen(argv[2], "w");
   if (!logfp)
   {
      printf("Cant create log file %s, error=%d\n",
              argv[2],errno);
      return 1;
   }
   /* line buffering only */
   setvbuf(logfp, (char *)NULL, _IOLBF, 0);

   act.sa_handler = sigCatch;
   sigemptyset(&act.sa_mask);
   act.sa_flags = SA_RESTART;
   if (sigaction(SIGTERM, &act, 0) != 0)
   {
      traceit("sigaction-TERM failed, error=%d\n", errno);
      return 1;
   }
   if (sigaction(SIGHUP, &act, 0) != 0)
   {
      traceit("sigaction-HUP failed, error=%d\n", errno);
      return 1;
   }
   if (sigaction(SIGQUIT, &act, 0) != 0)
   {
      traceit("sigaction-QUIT failed, error=%d\n", errno);
      return 1;
   }
   if (sigaction(SIGINT, &act, 0) != 0)
   {
      traceit("sigaction-INT failed, error=%d\n", errno);
      return 1;
   }

   do
   {
      /* process configuration file */
      rc = read_config(argv[1]);
      if (rc != 0)
      {
         traceit("Failed to process config file %s\n", argv[1]);
         break;
      }

      /* Open DB */ 
      rc = open_users(USERS);
      if (rc != 0)
      {
         traceit("Failed to open %s\n", USERS);
         break;
      }

      /* get ready to accept incoming connections */
      rc = srv_open();
      if (rc != 0)
      {
         traceit("srv_open() failed\n");
         break;
      }

      rc = internal_raw_open();
      if (rc != 0)
      {
         traceit("internal_raw_open() failed\n");
         break;
      }

      rc = external_raw_open();
      if (rc != 0)
      {
         traceit("external_raw_open() failed\n");
         break;
      }

      /* get ready to accept shell commands from netcat */
      rc = cmd_open();
      if (rc != 0)
      {
         traceit("cmd_open() failed\n");
         break;
      }
      traceit("dextra_srv %s initialized...entering processing loop\n", VERSION);

      /* Initialize our local repeater modules */
      for (i = 0; i < 3; i++)
         memset(&mod_links[i], '\0', sizeof(mod_links[i]));

      if (strlen(LINK_ON_STARTUP) == CALL_SIZE)
      {
         if (LINK_ON_STARTUP[0] == 'A')
            i = 0;
         else
         if (LINK_ON_STARTUP[0] == 'B')
            i = 1;
         else
         if (LINK_ON_STARTUP[0] == 'C')
            i = 2;
         else
            i = -1;
 
         if (i >= 0)
         {
            mod_links[i].from_mod = LINK_ON_STARTUP[0];
            memcpy(mod_links[i].to_call, LINK_ON_STARTUP + 1, 6);
            mod_links[i].to_call[6] = ' ';
            mod_links[i].to_call[7] = ' ';
            mod_links[i].to_call[CALL_SIZE] = '\0';
            mod_links[i].to_mod = LINK_ON_STARTUP[7];
            if (!link_command(i,true))
            {
               if (ANNOUNCE_LINK)
               {
                  snprintf(ANNOUNCE_NOTIFY_FILE, FILENAME_MAX, "%s/%c_failed_linked.dat",
                           ANNOUNCE_DIR_NOTIFY, LINK_ON_STARTUP[0]);
                  post_notification(ANNOUNCE_NOTIFY_FILE);
               }
               traceit("Failed to send link command\n");
            }
         }
      }
      print_links_file();
      runit();
      traceit("Leaving processing loop...\n"); 

   } while (false);

   /* Close all handles */
   if (cmd_sock != -1)
   {
      close(cmd_sock);
      traceit("Closed COMMAND_PORT\n");
   }

   if (internal_raw_sock != -1)
   {
      close(internal_raw_sock);
      traceit("Closed repeater-controller\n");
   }

   if (external_raw_sock != -1)
   {
      close(external_raw_sock);
      traceit("Closed g2\n");
   }

   if (srv_sock != -1)
   { 
      close(srv_sock);
      traceit("Closed LISTEN_PORT\n");
   }

   /* In case we have any links to our repeater modules, UN-LINK them */
   for (i = 0; i < 3; i++)
   {
      if (mod_links[i].to_mod != '\0')
         unlink_command(i);
   }

   print_links_file();
  
   /* Close database */
   if (dbp != NULL)
   {
      dbp->close(dbp, 0);
      traceit("Closed database %s\n",USERS);
   }
 
   traceit("dextra_srv exiting\n");
   fclose(logfp);
 
   return rc;
}

