#!/bin/bash

# Change ABCDEF to your G2 Gateway callsign
# It should NOT be more that 6 characters
REPEATER_CALL=ABCDEF

# The IP address assigned to the G2 EXTERNAL ethernet interface(on most G2 systems is eth0)
# eth0 is the G2 EXTERNAL ethernet interface
# If your G2 external ethernet interface is eth1, then use the IP address that is assigned to eth1
# This IP address is usually 10.x.x.x or 192.168.x.x
#    Run the command: ifconfig
#     or the command: ip addr show
#    to find out which is the correct IP to use. 
# Note that 127.0.0.1 may NOT work on some G2 machines
G2_IP=127.0.0.1

# How long to delay in dextra_detect_notification
#    before processing an EMPTY notification file in /root/dextra/tmp/
#    to send back a reply.
#    A delay of 5 seconds or more is sufficient.
#    A delay is required to give a few seconds to the repeater
#    to initialize itself.
DELAY_BEFORE=5

# Nothing to explain here
G2_PORT=40000

# 18 or 19 or 20 milliseconds is OK
# This is used as a normal delay between each DV packet
DELAY=20

# The directory where the announcement *.dat files are located.
#   These *.dat files are dvtool format files:
#         already_linked.dat
#         already_unlinked.dat
#         failed_linked.dat
#         linked.dat
#         unlinked.dat
#  The above files contain "link, unlink,..." announcements when 
#     your G2 is linked or unlinked or failed to link 
DAT_FOLDER=/root/dextra

# The directory where the empty notification files are created by dextra_srv
# This directory must exist and it is the same directory that you will find in dextra_srv.cfg
ANNOUNCE_DIR_NOTIFY=/root/dextra/tmp

# The directory where dextra is installed. All the binaries are there
DEXTRA_BIN_DIR=/root/dextra

# stop any old DETECTION programs running
killall dextra_detect_notification  >/dev/null 2>&1
sleep 1

# just in case it hangs
killall -s SIGKILL dextra_detect_notification >/dev/null  2>&1
sleep 1

# start the detection
/bin/rm -rf ${ANNOUNCE_DIR_NOTIFY}/*
${DEXTRA_BIN_DIR}/dextra_detect_notification ${DEXTRA_BIN_DIR} $G2_IP $G2_PORT $DAT_FOLDER $REPEATER_CALL $DELAY $ANNOUNCE_DIR_NOTIFY $DELAY_BEFORE >/var/log/dextra_detect_notification.log 2>&1 &

exit 0

