#!/bin/bash

NOW=`date`
echo $0 Starting at $NOW

# Inside dextra_srv.cfg, you will see a value for COMMAND_PORT
#  Use the same value here.
COMMAND_PORT=30003

PSQL=`which --skip-alias -- psql`
RM=`which --skip-alias -- rm`
CAT=`which --skip-alias -- cat`
NC=`which --skip-alias -- nc`

if [ -z $PSQL ] ; then
	echo Could not locate psql
	exit 1
fi

if [ -z $RM ] ; then
        echo Could not locate rm
        exit 1
fi

if [ -z $CAT ] ; then
        echo Could not locate cat
        exit 1
fi

if [ -z $NC ] ; then
        echo Could not locate nc
        exit 1
fi

$RM -rf /tmp/users.txt
$RM -rf /tmp/xrfdata.dat
$RM -rf /tmp/sync_gip.dat
$RM -rf /tmp/sync_mng.dat

$PSQL -d dstar_global -U postgres << EOF 
COPY xrfdata (xrf_name, xrf_domain) to '/tmp/xrfdata.dat' using delimiters '|';
COPY sync_gip (zonerp_cs, zonerp_ipaddr) to '/tmp/sync_gip.dat' using delimiters '|';
COPY sync_mng(target_cs) to '/tmp/sync_mng.dat' using delimiters '|';
\q
EOF

if [ -s /tmp/xrfdata.dat ] ; then

	$CAT /tmp/xrfdata.dat  >>  /tmp/users.txt
 	$CAT /tmp/sync_gip.dat >>  /tmp/users.txt
	$CAT /tmp/sync_mng.dat >>  /tmp/users.txt

	$RM -rf /root/dextra/dextra.db_new
	/root/dextra/dextra_create_db /root/dextra/dextra.db_new /tmp/users.txt 
	RC=$?
	if [ $RC -eq 0 ] ; then

		echo Executing  netcat command with PORT $COMMAND_PORT

		$NC -u -w 1 127.0.0.1 $COMMAND_PORT  << EOF
upd /root/dextra/dextra.db_new
EOF
	else
                echo Return code $RC failed to create the new file dextra.db_new
	fi
else
	echo "File /tmp/xrfdata.dat not extracted from database" 
fi 

NOW=`date`
echo $0 Finished at $NOW
